/* ---------------------------------- NETFIND ----------------------------------------- *
 *
 * Pure Data external to provide the netsendv3~ and netreceivev3 objects a
 * device discovery service to connect to other such objects on other PD platforms
 * on the same LAN or WLAN. Please note: the left outlet must be connected to a netreceivev3
 * and the right outlet to a netsendv3 object.
 * 
 * Reference paper for the SABy protocol: 
 * L. Gabrielli, S.Squartini, F. Piazza, “Advancements and Performance analysis on the 
 * Wireless Music Studio (WeMUST) framework”, in Proc. AES 134th Convention, Rome, Italy, 2013
 *
 * Written by Leonardo Gabrielli <l.gabrielli@univpm.it>
 * Source and updates at http://a3lab.dii.univpm/wemust/netfind
 *										
 * Copyright (C) 2013 L.Gabrielli						
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * See file LICENSE for further informations on licensing terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License            
 * along with this program; If not, see <http://www.gnu.org/licenses/> 		
 *
 * Based on PureData by Miller Puckette and others.
 *
 */



#ifdef _WIN32
#else
#define _GNU_SOURCE // specifies the use of GNU specific functions as strcasestr,strchrnul,strtof
#endif

#include "m_pd.h"
#include "s_stuff.h"

#include <sys/types.h>
#include <string.h>
#include <math.h>
#ifdef _WIN32
#include <winsock.h>
#include "pthread.h"
#else
#include <errno.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <time.h>
#include <unistd.h>
#include <signal.h>
#define SOCKET_ERROR -1
#endif



/* --- Custom library for sparing some TCP hassle --- */
#include "libmsock/msock.h"



/* --- PROTOCOL DEFINES --- */
#define SABY_GROUP			"239.255.255.251"
#define SABY_PORT			1901
#define FLOWCTRL_TCP_PORT		(unsigned int) 8022 // free
#define REPLY_TCP_PORT			(unsigned int) 8026 // free
#define ANNOUNCE_INTERVAL		20 // seconds
#define MAX_CONNECTED_PEERS		1
#define MAX_NAME_LEN			10
#define MAX_MESSAGE_LEN			256
#define MAX_TAGSTRING_LEN		64

#define __testSABy "SABy-"
#define __testADV "ADV"
#define __testSRC "SRC"
#define __testREP "REP"
#define __SEARCH_ALL_WILDCARD "*"

#define NEGOTIATION_START		"NEGOSTART\n\r"
#define NEGOTIATION_ACK			"NEGOACK\n\r"
#define NEGOTIATION_OK			"NEGOOK\n\r"
#define NEGOTIATION_FAIL		"NEGOFAIL\n\r" // when parameters do not match (e.g. different sample rates SR)
#define FLOWCTRL_CHECK			"FC_CH\n\r"
#define FLOWCTRL_ACK			"FC_ACK\n\r"
#define FLOWCTRL_INTERVAL		5
#define SOCKET_RETRY_INTERVAL		5
#define MSGLINES			5


/* --- AUX CLASSES --- */
typedef struct _announce
{
	char 						appField[20]; 			// "puredata"
	t_float 					SR; 				// sample rate
	int						framesize;			// frame size
	char 						name[MAX_NAME_LEN+1];		// user defined name o unique ID
	int 						connected;			// connection with peer status
	struct in_addr					addr;				// address
	in_port_t 					port;				// port
	int						sock_fd;			// when this struct is used as a storage for peers here I put the socket for the connection with him
	char						tags[MAX_TAGSTRING_LEN+1];	// tags string
	char* 						tags_ptr;			// position of the write head, useful to update the tags with addtag
	// if goint into jack: channels, period...
} t_announce;


/* --- PUREDATA EXTERNAL CLASS --- */
typedef struct _netfind
{
    t_object 					x_obj;
    t_outlet* 					x_nrcv_outlet;			// outlet to control a netreceive object
    t_outlet*					x_nsend_outlet; 		// outlet to control a netsend object

    int 					x_broadPortNo; 			// multicast port
    int 					x_connecState;			// 0 if no connection, otherwise the number of connections open (those running flowctrl)

    int 					x_read_multicastSocket; 	// socket fd for listen thread
    int 					x_write_multicastSocket;	// socket fd for announce/search
    struct sockaddr_in*				x_write_dest_addr;		// multicast sockaddr_in struct (for announce/search)
    int 					x_write_addrlen;		// multicast addrlen info (for announce/search)
    int						x_accepted_sock_fd[MAX_CONNECTED_PEERS];		// liste_flow_ctrl socket array of accepted peers
    int 					x_listen_sock_fd;		// listen_flow_ctrl socket
    int						x_search_REPLY_sock;		// socket for listening on REPLY_TCP_PORT
    int 					x_accepted;			// number of accepted peers, up to MAX_CONNECTED_PEERS

    t_atom*					x_tl_ptr; 			// tag list ptr (to be used when tags are added as a list via [addtags(
    int						x_tl_sz; 			// size of the tag list
    t_symbol*					x_s; 				// temp symbol ptr for addtags
    char					x_tag_delimiter;
    t_announce 					x_personalData;			// pointer to announce class to store personal info
    t_announce 					x_peers[MAX_CONNECTED_PEERS];	// pointer to announce classes to store peers info
    int 					x_cp;				// current peer number, used only for passing arguments to functions

    char					x_searchTag[MAX_TAGSTRING_LEN];

    pthread_mutex_t				x_mutex;
    pthread_cond_t				x_requestcondition;
    pthread_cond_t				x_answercondition;
    pthread_t					x_listen_thread,		// listen: listens to netfind announces
    						x_announce_thread, 		// announce: announce himself periodically on the netfind
    						x_connect_thread, 		// connect: creates a TCP connection for flow control of the netsend patch we are connected to
    						x_flow_ctrl_thread,		// flow control thread
    						x_flow_ctrl_listen_thread,	// flow control listen thread
    						x_netfind_search_thread;	// issues a search and waits for replies
    										// NOTE: announce<-->listen, flow_ctrl<-->flow_ctrl_listen
} t_netfind;

/* --- CLASS ALLOC --- */
static t_class *netfind_class;



/* --- FUNCTION DECLARATIONS --- */
static void*		netfind_listen(void *xx);
static void*		netfind_announce(void *xx);
static void 		netfind_flow_ctrl(t_netfind*);
static void 		netfind_flow_ctrl_listen(t_netfind*);
static void		netfind_taglist(t_netfind*,t_symbol*,int,t_atom*);
static void		netfind_addtag(t_netfind*, t_symbol*);
static void 		netfind_search(t_netfind *, t_symbol *);
int			parseAnnounceMsg(t_netfind*, int, char*);
static int 		myServerSocket(short port);
static int 		myClientSocket(struct in_addr addr, short port);
char*			strip_chars(const char *string, const char *chars);
int			isNewPeer(t_netfind* x, int cp);
void			UUID_gen(char *s, const int len);
int			send_SABy_msg(t_netfind*, const char*, int);
void			netfind_do_search(t_netfind*);


/* --- PUREDATA: _NEW METHOD --- */
static void *netfind_new(t_symbol* input_str) //, t_symbol* stringa per match buddy
{

	// PD OBJECT INIT
	t_netfind *x = (t_netfind *)pd_new(netfind_class);
	x->x_nrcv_outlet = outlet_new(&x->x_obj, &s_symbol);
	x->x_nsend_outlet = outlet_new(&x->x_obj, &s_symbol);


	// CONNECTION INITs
	x->x_read_multicastSocket = -1;
	x->x_write_multicastSocket = -1;
	x->x_broadPortNo = SABY_PORT; 
	x->x_connecState = 0;
	x->x_listen_thread = 0;
	x->x_announce_thread = 0;
	x->x_connect_thread = 0;
	x->x_flow_ctrl_thread = 0;
	x->x_flow_ctrl_listen_thread = 0;
	x->x_netfind_search_thread = 0;

	x->x_tag_delimiter = 0x2c; // comma

	// PERSONAL DATA INITs
	if( (int)strlen(input_str->s_name) > MAX_NAME_LEN) {
		error("netfind: warning: name (1st arg) too long, clipped to %dchar",MAX_NAME_LEN-1);
	input_str->s_name[MAX_NAME_LEN] = '\0';
	}

	if( (int)strlen(input_str->s_name) == 0 ) {
		UUID_gen(x->x_personalData.name,MAX_NAME_LEN-1); // generate random alphanumeric name
		post("netfind: your name is: %s",x->x_personalData.name);
	}
	else
		strncpy(x->x_personalData.name,input_str->s_name,MAX_NAME_LEN);

	strcpy(x->x_personalData.appField,"puredata\0");
	x->x_personalData.tags_ptr = &x->x_personalData.tags;

	strcpy(x->x_searchTag,__SEARCH_ALL_WILDCARD); // accept any peer

	x->x_personalData.SR = (int)sys_getsr();
	x->x_personalData.framesize = sys_getblksize();

	x->x_personalData.port = 0;
	struct in_addr zero = {0};
	x->x_personalData.addr = zero; // ToDo using int getifaddrs(struct ifaddrs **ifap);

	/* START CHILD THREADS */
	if( pthread_create(&x->x_announce_thread, 0, netfind_announce, x) != 0) {
		error("error creating send thread");
	}

	sleep(1);


	if( pthread_create(&x->x_listen_thread, 0, netfind_listen, x) != 0) {
		error("error creating listen thread");
	}

	sleep(1);

	if( pthread_create(&x->x_flow_ctrl_listen_thread, 0, netfind_flow_ctrl_listen, x) != 0) {
		error("error creating flow_ctrl listen thread");
	}

	return (x);
}




/* --- CHILD THREAD TO LISTEN FROM MULTICAST GROUP --- */
static void *netfind_listen(void *xx) {

	t_netfind *x = (t_netfind *)xx;
	struct sockaddr_in remote_addr;
	int cnt;
	struct ip_mreq mreq;
	int cp = 0; // counter for connected peers
	int ret;
	unsigned int i, msglen;
	int broadSocket;

	struct msghdr aMsg; 
	struct cmsghdr *ctrl_data;
	char in_buf[MAX_MESSAGE_LEN];
	char parseBuf[MAX_MESSAGE_LEN];

	aMsg.msg_name       =	getbytes(sizeof(struct sockaddr_in));
	aMsg.msg_namelen    =	sizeof(struct sockaddr);
	aMsg.msg_iov        =	getbytes(sizeof(struct iovec));
	aMsg.msg_iovlen     =	1;
	aMsg.msg_iov->iov_base = in_buf;
	aMsg.msg_iov->iov_len =	sizeof(in_buf)-1;
	aMsg.msg_control    =	getbytes(sizeof(char)*100);
	aMsg.msg_controllen =	sizeof(struct cmsghdr)+4+18;/*4 byte ipaddr + 18 TRRI*/

listen_top:

	/* set up socket */
	broadSocket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);

	while (broadSocket < 0) {
		post("listen socket opening fail, will retry in %d secs\n",SOCKET_RETRY_INTERVAL);
		sleep(SOCKET_RETRY_INTERVAL);
		broadSocket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	}

	bzero((char *)&remote_addr, sizeof(remote_addr));
	remote_addr.sin_family = AF_INET;
	remote_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	remote_addr.sin_port = htons(SABY_PORT);

	while(bind(broadSocket, (struct sockaddr *) &remote_addr, sizeof(remote_addr)) != 0) {
		post("netfind_listen: socket bind fail on port %d, with error %d, will retry in %d secs\n",(int)SABY_PORT,(int)errno,SOCKET_RETRY_INTERVAL);
		sleep(SOCKET_RETRY_INTERVAL); // wait 5 seconds and retry
	}

	mreq.imr_multiaddr.s_addr = inet_addr(SABY_GROUP);
	mreq.imr_interface.s_addr = htonl(INADDR_ANY);

	if (setsockopt(broadSocket, IPPROTO_IP, IP_ADD_MEMBERSHIP,&mreq, sizeof(mreq)) < 0) 
		error("setsockopt mreq fail");

	pthread_mutex_lock(&x->x_mutex);
	x->x_read_multicastSocket = broadSocket;
	pthread_mutex_unlock(&x->x_mutex);

	while (1) {
listen_while:
		cnt = recvmsg(broadSocket, &aMsg, 0);
		if (cnt < 0) {
			break;
		}
		else if (cnt == 0) {
			break;
		}
		else {
			strncpy(parseBuf,aMsg.msg_iov->iov_base,MAX_MESSAGE_LEN-1);
			if( strcasestr(parseBuf, __testSABy) == NULL ) { 
				error("invalid puredata netfind message received");
			} else { 
				if( strcasestr(parseBuf, __testADV) != NULL ) {
					ret = parseAnnounceMsg(x, cp, parseBuf); //  parsing
					if( ret < 0 ) 
						goto listen_while;
					
					if( strstr(x->x_searchTag,__SEARCH_ALL_WILDCARD) == NULL ) {
						if( strcasestr(x->x_searchTag,x->x_peers[cp].tags) == NULL)
							goto listen_while;
					}
					pthread_mutex_lock(&x->x_mutex);
					struct sockaddr_in *src = aMsg.msg_name;
					x->x_peers[cp].addr.s_addr = (src->sin_addr.s_addr);
					x->x_peers[cp].port = ntohs(src->sin_port);
					pthread_mutex_unlock(&x->x_mutex);
					if ( isNewPeer(x,cp) == 1 ) {
						pthread_mutex_lock(&x->x_mutex);
						x->x_cp = cp;
						// unlocking will be done only after creation to preserve cp
						if( pthread_create(&x->x_flow_ctrl_thread, 0, netfind_flow_ctrl, x) != 0) {
							error("error creating connection thread\n");
						}
						if( ++cp >= MAX_CONNECTED_PEERS ) //advance so that we can later connect to a new peer --> NOT IMPLEMENTED YET!
							goto max_connections;
					}
				}
				else if( strcasestr(parseBuf, __testSRC) != NULL ) {

					ret = parseAnnounceMsg(x, cp, parseBuf); 
					if( ret < 0 ) 
						goto listen_while;				
					//if the issuer looks for a tag and I do not own the same tag DO NOT reply
					if( strstr(x->x_peers[cp].tags,__SEARCH_ALL_WILDCARD) == NULL ) {
						if( strcasestr(x->x_personalData.tags,x->x_peers[cp].tags) == NULL)
							goto listen_while;
					}

					// address of the search issuer
					struct sockaddr_in *src = aMsg.msg_name;
					x->x_peers[cp].addr.s_addr = (src->sin_addr.s_addr); 
					x->x_peers[cp].port = ntohs(src->sin_port);

					cnt = send_SABy_msg(x,__testREP,cp);

				}
				else if( strcasestr(parseBuf, __testREP) != NULL ) {;} // donothing. A reply to the multicast group is not accepted
			}
		}
	}

max_connections:
	post("netfind: max number of peer reached... will not accept connections\n");
	while(cp >= MAX_CONNECTED_PEERS)
		sleep(10);
	goto listen_top;

	return;
}








/* --- CHILD THREAD TO ANNOUNCE OVER SSDP --- */
static void *netfind_announce(void *xx) {

	t_netfind *x = (t_netfind *)xx;
	struct sockaddr_in dest_addr;
	int addrlen, sock, cnt;
	int broadSocket;

	/* set up socket */
	broadSocket = socket(AF_INET, SOCK_DGRAM, 0);

	while (broadSocket < 0) {
		post("announce socket opening fail, will retry in %d secs\n",SOCKET_RETRY_INTERVAL);
		sleep(SOCKET_RETRY_INTERVAL);
		broadSocket = socket(AF_INET, SOCK_DGRAM, 0);
	}


	bzero((char *)&dest_addr, sizeof(dest_addr));
	dest_addr.sin_family = AF_INET;
	dest_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	dest_addr.sin_port = htons(SABY_PORT);
	addrlen = sizeof(dest_addr);

	dest_addr.sin_addr.s_addr = inet_addr(SABY_GROUP);


	pthread_mutex_lock(&x->x_mutex);
	x->x_write_multicastSocket = broadSocket;
	x->x_write_dest_addr = &dest_addr;
	x->x_write_addrlen = addrlen;
	pthread_mutex_unlock(&x->x_mutex);

	while (1) {
		cnt = send_SABy_msg(x,__testADV,0);
		if (cnt < 0) {
			post("netfind_announce: sendto: error %d, will retry in %d secs...\n",(int)errno,SOCKET_RETRY_INTERVAL);
			sleep(SOCKET_RETRY_INTERVAL);
		}
		else {
			sleep(ANNOUNCE_INTERVAL); // sleep some seconds
		}
	}
	return;
}





/* search for a tag */
static void netfind_search(t_netfind *x, t_symbol *searchTag) {

	strncpy(x->x_searchTag,searchTag->s_name,MAX_TAGSTRING_LEN); // store the search tag for the thread to be called

	if( x->x_netfind_search_thread ) {
		pthread_cancel(x->x_netfind_search_thread);
		close(x->x_search_REPLY_sock);
	}

	if( pthread_create(&x->x_netfind_search_thread, 0, netfind_do_search, x) != 0) {
		error("error creating send thread");
	}
}

/* thread that actually implements the real search. ToDO: add a timer to stop searching after a while */
void netfind_do_search(t_netfind* x) {

	int cnt = 0;
	int listen_fd;
	char* rcvmsg = getbytes(sizeof(char)*MAX_MESSAGE_LEN);
	struct sockaddr_in cli_addr;

	while(!x->x_write_multicastSocket) {
		sleep(SOCKET_RETRY_INTERVAL); // wait for the socket to be created by another thread
	}

	cnt = send_SABy_msg(x,__testSRC,0);
	if( cnt < 0) {;}

	listen_fd = myServerSocket(REPLY_TCP_PORT);

	pthread_mutex_lock(&x->x_mutex);
	x->x_search_REPLY_sock = listen_fd;
	pthread_mutex_unlock(&x->x_mutex);

	if( listen_fd == -1 )
		error("ServerSocket error\n");

	listen(listen_fd,0); // ToDo: handle multiple peers
	int clilen = sizeof(cli_addr);

	int newsock_fd = accept(listen_fd,(struct sockaddr *) &cli_addr,&clilen);
	if (newsock_fd < 0)
	    	error("ERROR on accept");

	cnt = sockGets(newsock_fd,rcvmsg,MAX_MESSAGE_LEN-1);
	if( cnt < 0 ) {
		error("connection probably closed during sockGets");
	}
	close(newsock_fd);

	pthread_mutex_lock(&x->x_mutex);
	int cp = x->x_cp + 1;
	pthread_mutex_unlock(&x->x_mutex);

	if( strcasestr(rcvmsg, __testSABy) == NULL ) {
		post("invalid puredata netfind message");
	} 
	else {
		if( strcasestr(rcvmsg, __testADV) != NULL ) {
			int ret = parseAnnounceMsg(x, cp, rcvmsg);
			if( ret < 0 )
				return;
			post("netfind: peer data: app: %s, name = %s, tags: %s, SR = %d, framesize: %d \n", x->x_peers[cp].appField, x->x_peers[cp].name, x->x_peers[cp].tags, (int)x->x_peers[cp].SR, (int)x->x_peers[cp].framesize);

			pthread_mutex_lock(&x->x_mutex); 
			x->x_peers[cp].addr.s_addr = cli_addr.sin_addr.s_addr;  
			x->x_peers[cp].port = ntohs(cli_addr.sin_port);
	
			if ( isNewPeer(x,cp) == 1 ) {
				x->x_cp = cp; 
				if( pthread_create(&x->x_flow_ctrl_thread, 0, (void*)netfind_flow_ctrl, x) != 0) {
					error("error creating connection thread\n");
					pthread_mutex_unlock(&x->x_mutex);
				}
	
				if( ++cp >= MAX_CONNECTED_PEERS ) { //advance so that we can later connect to a new peer --> ToDo
					error("Max number of connections reached");
					return;
				}
			}
		}
	}
}
// {}


static void netfind_flow_ctrl_listen(t_netfind* x) {

	int cnt, ret;
	char* rcvstr = getbytes(sizeof(char)*MAX_MESSAGE_LEN);
	char tmpstr[9];
	int listen_fd;
	int flowctrl_connected;
	struct timeval *tv = getbytes(sizeof(struct timeval));
	struct sockaddr_in cli_addr;
	char *negotiation_start = strip_chars(NEGOTIATION_START, "\n\r");
	char *flowctrl_check = strip_chars(FLOWCTRL_CHECK, "\n\r");
	int peer_nrcv_port, nrcv_port, nsend_port;
	int rcvsock_fd;
	struct sockaddr_in portneg_addr;
	char nsend_msg[30]; 
	char nrcv_msg[11]; 

	char szclient_host[64],szclient_ip[64];
	u_short client_port;

wait_connection: 
	listen_fd = myServerSocket(FLOWCTRL_TCP_PORT); 

	if( listen_fd == -1 )
		error("ServerSocket error\n");

	x->x_listen_sock_fd = listen_fd;

	listen(listen_fd,0); // WARNING: only 1 connection is handled for now
    int clilen = sizeof(cli_addr);
    int newsock_fd = accept(listen_fd,(struct sockaddr *) &cli_addr,
	                 &clilen);
    if (newsock_fd < 0)
    	error("ERROR on accept");
	cnt = sockGets(newsock_fd,rcvstr,MAX_MESSAGE_LEN-1);
	if( cnt < 0 ) { 
		error("connection probably closed during sockGets");
		sleep(5);
		close(newsock_fd);
		goto wait_connection;
	}

	if( strstr(rcvstr,negotiation_start) == NULL ) { 
		error("START error");
		close(newsock_fd);
		goto wait_connection;
	}

	/* send ACK */
	if( cnt = sockPuts(newsock_fd,NEGOTIATION_ACK) < 0 ) {
		error("could not send ACK");
		close(newsock_fd);
		goto wait_connection;
	}

	if( cnt = sockGets(newsock_fd,rcvstr,MAX_MESSAGE_LEN-1) < 0 ) { 
		error("connection probably closed during sockGets");
		close(newsock_fd);
		goto wait_connection;
	}
	sprintf(tmpstr, "%d\0",(int)x->x_personalData.SR);
	if( strstr(rcvstr,tmpstr) == NULL ) { // SR check
		sockPuts(newsock_fd,NEGOTIATION_FAIL);
		close(newsock_fd);
		error("SR do not match, cannot connect");
		goto wait_connection;
	} else {
		if( cnt = sockPuts(newsock_fd,NEGOTIATION_OK) < 0 ) {
			error("could not send OK");
			close(newsock_fd);
			goto wait_connection;
		}
	}


	/* port negotiation */
port_negotiation:
	if( cnt = sockGets(newsock_fd,rcvstr,MAX_MESSAGE_LEN-1) < 0 ) {
		error("connection probably closed during sockGets");
		close(newsock_fd);
		goto wait_connection;
	}
	peer_nrcv_port = atoi(rcvstr); 
	nrcv_port = peer_nrcv_port + 1;
	/* check if this port is free */
	rcvsock_fd = socket(AF_INET, SOCK_STREAM, 0);
	if( rcvsock_fd < 0 ) {
		error("netfind: port negotiation socket error\n");
		if( sockPuts(newsock_fd,NEGOTIATION_FAIL) < 0) {
			close(newsock_fd);
			goto wait_connection;
		}
		goto port_negotiation;
	}
	bzero((char *) &portneg_addr, sizeof(portneg_addr));
	portneg_addr.sin_family = AF_INET;
	portneg_addr.sin_addr.s_addr = INADDR_ANY;
	portneg_addr.sin_port = htons(nrcv_port);
	if (bind(rcvsock_fd, (struct sockaddr *) &portneg_addr, sizeof(portneg_addr)) < 0) {
		if( errno == EADDRINUSE ) { // try a new port
			if( sockPuts(newsock_fd,NEGOTIATION_FAIL) < 0) { 
				close(newsock_fd);
				goto wait_connection;
			}
			goto port_negotiation;
		} else { // retry
			if( sockPuts(newsock_fd,NEGOTIATION_FAIL) < 0) {
				close(newsock_fd);
				goto wait_connection;
			}
			goto port_negotiation;
		}
	}
    if (close (rcvsock_fd) < 0 ) {
        if( sockPuts(newsock_fd,NEGOTIATION_FAIL) < 0) { 
			close(newsock_fd);
			goto wait_connection;
		}
		goto port_negotiation;
    }

	if( cnt = sockPuts(newsock_fd,NEGOTIATION_OK) < 0 ) {
		error("could not send OK");
		close(newsock_fd);
		goto wait_connection;
	}

	/* PORT NEGOTIATION OK, initialize netsend and netreceive */
	x->x_accepted_sock_fd[x->x_accepted++] = newsock_fd;
	nsend_port = peer_nrcv_port;

	t_atom *port_a = getbytes(sizeof(t_atom));
	SETFLOAT(port_a,nrcv_port);
	outlet_anything(x->x_nrcv_outlet,gensym("port"),1,port_a);

	t_atom connect_values[2];
	SETSYMBOL(&connect_values[0],gensym(inet_ntoa(cli_addr.sin_addr)));
	SETFLOAT(&connect_values[1],nsend_port);
	outlet_anything(x->x_nsend_outlet,gensym("connect"),2,connect_values);

	/* negotiation finished, now we control the flow */
	flowctrl_connected = 1;

	while(flowctrl_connected == 1) {
		if( cnt = sockGets(newsock_fd,rcvstr,MAX_MESSAGE_LEN-1) < 0 ) { 
			error("peer died?");
			flowctrl_connected = 0;
			return;
		}

		if( strstr(rcvstr,flowctrl_check) == NULL ) { // ACK check
			error("ACK error: peer in bad state? will ignore");
		}

		if( cnt = sockPuts(newsock_fd,FLOWCTRL_ACK) < 0 ) {
			error("could not check peer state");
			flowctrl_connected = 0;
			return;
		}

	}

	return;
}








/* flow control TCP connection */
static void netfind_flow_ctrl(t_netfind* x) {

	int cp = x->x_cp;
	pthread_mutex_unlock(&x->x_mutex);

	int cnt = 0;
	char* rcvstr = getbytes(sizeof(char)*MAX_MESSAGE_LEN);
	char message[MAX_MESSAGE_LEN];
	char nsend_msg[30]; 
	char nrcv_msg[11]; 
	unsigned int nsend_port,nrcv_port;
	char *negotiation_ack = strip_chars(NEGOTIATION_ACK, "\n\r");
	char *negotiation_ok = strip_chars(NEGOTIATION_OK, "\n\r");
	char *flowctrl_ack = strip_chars(FLOWCTRL_ACK, "\n\r");
	struct sockaddr_in portneg_addr;
	long elapsed;
	struct timeval t0, t1;
	time_t curtime;

	/* get initial timestamp for elapsed time calc */
	gettimeofday(&t0, NULL);
	curtime=t0.tv_sec;

	int peer_fd = myClientSocket(x->x_peers[cp].addr,FLOWCTRL_TCP_PORT);
	if( peer_fd == -1 ) {
		error("could not connect to peer on port %d",FLOWCTRL_TCP_PORT);
		x->x_peers[cp].connected = 0;
		pthread_mutex_lock(&x->x_mutex);
		x->x_cp--;
		pthread_mutex_unlock(&x->x_mutex);
		return;
	}

	/* first ACK */
	cnt = sockPuts(peer_fd,NEGOTIATION_START);
	if( cnt < 0 ) {
		error("netfind: could not start negotiation with peer");
		x->x_peers[cp].connected = 0;
		pthread_mutex_lock(&x->x_mutex);
		x->x_cp--;
		pthread_mutex_unlock(&x->x_mutex);
		return;
	}

	if( cnt = sockGets(peer_fd,rcvstr,MAX_MESSAGE_LEN-1) < 0 ) { 
		error("netfind: connection probably closed during sockGets");
		pthread_mutex_lock(&x->x_mutex);
		x->x_cp--;
		pthread_mutex_unlock(&x->x_mutex);
		x->x_peers[cp].connected = 0;
		return;
	}

	if( strstr(rcvstr,negotiation_ack) == NULL ) { // ACK check
		error("netfind: ACK error");
		x->x_peers[cp].connected = 0;
		pthread_mutex_lock(&x->x_mutex);
		x->x_cp--;
		pthread_mutex_unlock(&x->x_mutex);
		return;
	}

	/* negotiate the SR, redundant, but useful for future implementations */
	sprintf(message,"%d\n\r\0",	(int)x->x_personalData.SR);
	if( cnt = sockPuts(peer_fd,message) < 0 ) {
		error("netfind: could not negotiate SR with peer");
		x->x_peers[cp].connected = 0;
		pthread_mutex_lock(&x->x_mutex);
		x->x_cp--;
		pthread_mutex_unlock(&x->x_mutex);
		return;
	}

	if( cnt = sockGets(peer_fd,rcvstr,MAX_MESSAGE_LEN-1) < 0 ) { 
		error("netfind: connection probably closed during sockGets");
		x->x_peers[cp].connected = 0;
		pthread_mutex_lock(&x->x_mutex);
		x->x_cp--;
		pthread_mutex_unlock(&x->x_mutex);
		return;
	}

	if( strstr(rcvstr,negotiation_ok) == NULL ) { // SR OK?
		error("netfind: SR not accepted");
		x->x_peers[cp].connected = 0;
		pthread_mutex_lock(&x->x_mutex);
		x->x_cp--;
		pthread_mutex_unlock(&x->x_mutex);
		return;
	}

	/* PORT HANDSHAKING */
	if(RAND_MAX != 2147483647) {
		error("netfind: RAND_MAX is set to a different value on this machine. Check your architecture and edit the source. Abort.");
		x->x_peers[cp].connected = 0;
		pthread_mutex_lock(&x->x_mutex);
		x->x_cp--;
		pthread_mutex_unlock(&x->x_mutex);
		return;
	}

	/* negotiate the receive ports */
port_negotiation:
	srandom(time(NULL));
	nrcv_port = (unsigned int) (49152 + floor(random() / pow(2,17)) - 2);  // random port in the "dynamic range" [49152:65535]

	/* check if this port is free to bind on my system */
	int rcvsock_fd = socket(AF_INET, SOCK_STREAM, 0);
	if( rcvsock_fd < 0 )
	    goto port_negotiation;

	bzero((char *) &portneg_addr, sizeof(portneg_addr));
	portneg_addr.sin_family = AF_INET;
	portneg_addr.sin_addr.s_addr = INADDR_ANY;
	portneg_addr.sin_port = htons(nrcv_port);
	if (bind(rcvsock_fd, (struct sockaddr *) &portneg_addr, sizeof(portneg_addr)) < 0) {
		if( errno == EADDRINUSE ) {
			goto port_negotiation;
		}
	}

	if (close (rcvsock_fd) < 0 ) {
		goto port_negotiation;
    	}

	sprintf(message,"%d\n\r\0",nrcv_port);
	if( cnt = sockPuts(peer_fd,message) < 0 ) {
		error("netfind: could not negotiate port with peer");
		x->x_peers[cp].connected = 0;
		pthread_mutex_lock(&x->x_mutex);
		x->x_cp--;
		pthread_mutex_unlock(&x->x_mutex);
		return;
	}

	if( cnt = sockGets(peer_fd,rcvstr,MAX_MESSAGE_LEN-1) < 0 ) {
		error("netfind: connection probably closed during sockGets");
		x->x_peers[cp].connected = 0;
		pthread_mutex_lock(&x->x_mutex);
		x->x_cp--;
		pthread_mutex_unlock(&x->x_mutex);
		return;
	}

	if( strstr(rcvstr,negotiation_ok) == NULL ) { // SR OK?
		error("netfind: PORT not accepted");
		goto port_negotiation;
	}

	/* elapsed time for negotiation */
	gettimeofday(&t1, 0);
	elapsed = (t1.tv_sec-t0.tv_sec)*1000000 + t1.tv_usec-t0.tv_usec;
	//printf("DEBUG: Negotiation elapsed %ld us\n",elapsed);


	/* if we get here everything is fine: we can activate netsend and netreceive */
	nsend_port = nrcv_port + 1;
	x->x_peers[cp].connected = 1;
	x->x_connecState++;
	x->x_peers[cp].sock_fd = peer_fd;

	t_atom *port_a = getbytes(sizeof(t_atom));
	SETFLOAT(port_a,nrcv_port);
	outlet_anything(x->x_nrcv_outlet,gensym("port"),1,port_a);

	t_atom connect_values[2];
	SETSYMBOL(&connect_values[0],gensym(inet_ntoa(x->x_peers[cp].addr)));
	SETFLOAT(&connect_values[1],nsend_port);
	outlet_anything(x->x_nsend_outlet,gensym("connect"),2,connect_values);

	/* flow control */
	while(x->x_peers[cp].connected == 1) {

		if( cnt = sockPuts(peer_fd,FLOWCTRL_CHECK) < 0 ) {
			error("netfind: connection broke");
			x->x_peers[cp].connected = 0;
			pthread_mutex_lock(&x->x_mutex);
			x->x_cp--;
			pthread_mutex_unlock(&x->x_mutex);
			return;
		}

		if( cnt = sockGets(peer_fd,rcvstr,MAX_MESSAGE_LEN-1) < 0 ) { 
			error("netfind: peer died?");
			x->x_peers[cp].connected = 0;
			pthread_mutex_lock(&x->x_mutex);
			x->x_cp--;
			pthread_mutex_unlock(&x->x_mutex);
			return;
		}


		if( strstr(rcvstr,flowctrl_ack) == NULL ) { // ACK check
			error("netfind: ACK error: peer in bad state? will ignore");
		}

		sleep(FLOWCTRL_INTERVAL);
	}


	return;
}





static void netfind_addtag(t_netfind* x, t_symbol* s) {

	int freespace = MAX_TAGSTRING_LEN - (x->x_personalData.tags_ptr - x->x_personalData.tags);
	if( strlen(s->s_name) < freespace ) {
		sprintf(x->x_personalData.tags_ptr,"%s%c",s->s_name,x->x_tag_delimiter);
		x->x_personalData.tags_ptr += 1 + strlen(s->s_name);
	}
	else {
		error("netfind: tag string too long");
	}
}




/* add tags.  Warning: replaces possible existing tags */
static void netfind_taglist(t_netfind* x,t_symbol* s,int argc, t_atom* argv) { 

	t_atom* tl_ptr; 
	int j, i = argc, sym_len = 0;
	int tags_room = MAX_TAGSTRING_LEN; 
	char* wrhead = x->x_personalData.tags;

	if(x->x_tl_ptr){
		freebytes(x->x_tl_ptr, x->x_tl_sz * sizeof(t_atom));
		x->x_tl_ptr = 0;
	}
	x->x_s = s;
	x->x_tl_sz = argc;

	// brute force clear the string
	for(j = 0; j < MAX_TAGSTRING_LEN+1; j++)
		x->x_personalData.tags[j] = '\0';

	if(argc){
		x->x_tl_ptr=(t_atom*)getbytes(argc * sizeof(t_atom));
	}
	if(x->x_tl_ptr){
		tl_ptr = x->x_tl_ptr;
		while(argc--){
			if(A_SYMBOL==argv->a_type){
				sym_len = strlen(argv->a_w.w_symbol->s_name);
				if( (sym_len > 0) && (sym_len < tags_room) ) {
					sprintf(wrhead,"%s%c",argv->a_w.w_symbol->s_name,x->x_tag_delimiter);
					wrhead += sym_len+1;
					tags_room -= sym_len + 1;
				}
			}
			else {
				atom_string(argv, wrhead, MAX_TAGSTRING_LEN);
				wrhead = strchrnul(x->x_personalData.tags, '\0');
				sprintf(wrhead,"%c",x->x_tag_delimiter); wrhead++; 
			}
			argv++;
		} 
	}
#if DEBUG == 1
	else
		post("netfind: Could not allocate tag atoms");
#endif

}












/* --- FREE METHOD --- */

static void netfind_free(t_netfind *x)
{
	if( close(x->x_write_multicastSocket) != 0)
		error("netfind: close socket error %i",errno);
	if( close(x->x_read_multicastSocket) != 0)
			error("netfind: close socket error %i",errno);

	if( x->x_listen_thread ) {
		if( pthread_cancel(x->x_listen_thread) != 0)
			error("netfind: child thread did not cancel");
	}

	if( x->x_announce_thread ) {
		if( pthread_cancel(x->x_announce_thread) != 0)
			error("netfind: child thread did not cancel");
	}
	if( x->x_flow_ctrl_thread ) {
		if( pthread_cancel(x->x_flow_ctrl_thread) != 0)
			error("netfind: child thread did not cancel");
	}
	if( x->x_flow_ctrl_listen_thread ) {
		if( pthread_cancel(x->x_flow_ctrl_listen_thread) != 0)
			error("netfind: child thread did not cancel");
	}
	if( x->x_connect_thread ) {
		if( pthread_cancel(x->x_connect_thread) != 0)
			error("netfind: child thread did not cancel");
	}
	if( x->x_netfind_search_thread ) {
		if( pthread_cancel(x->x_netfind_search_thread) != 0)
			error("netfind: child thread did not cancel");
	}

	/* close all sockets */
	if( x->x_connecState ) {
		while(x->x_connecState--)
		close(x->x_peers[x->x_connecState].sock_fd);
	}
	close(x->x_read_multicastSocket);
	close(x->x_write_multicastSocket);
	while(x->x_cp--)
		close(x->x_accepted_sock_fd[x->x_cp]);
	close(x->x_listen_sock_fd);

	// ToDo free everything (lazy!!!)
}


void netfind_setup(void)
{
    netfind_class = class_new(gensym("netfind"), (t_newmethod)netfind_new,
        (t_method)netfind_free,
        sizeof(t_netfind), CLASS_DEFAULT, A_DEFSYMBOL, 0);
    class_addmethod(netfind_class, (t_method)netfind_search, gensym("search"), A_DEFSYMBOL, A_DEFSYMBOL, 0);
    class_addmethod(netfind_class, (t_method)netfind_addtag, gensym("addtag"), A_DEFSYMBOL, A_DEFSYMBOL, 0);
    class_addlist(netfind_class, (t_method)netfind_taglist);
}





/*---- SABY MESSAGING FUNCTION ----- */
int send_SABy_msg(t_netfind* x, const char* msgType, int cp) {

	char message[MAX_MESSAGE_LEN];

	if( strcasestr(msgType, __testADV) ) {
		sprintf(message,"SABy-%s\r\n"
				"APP:%s\r\n"
				"NID:%s\r\n"
				"TAG:%s\r\n"
				"SMP:%d|%d\r\n",
				msgType,
				x->x_personalData.appField,
				x->x_personalData.name,
				x->x_personalData.tags,
				(int)x->x_personalData.SR,(int)x->x_personalData.framesize);
		return sendto(x->x_write_multicastSocket, message, sizeof(message), 0,
				(struct sockaddr *) x->x_write_dest_addr, x->x_write_addrlen);
	}
	else if( strcasestr(msgType, __testSRC)) {
		sprintf(message,	"SABy-%s\r\n"
				"APP:\r\n"
				"NID:\r\n"
				"TAG:%s\r\n"
				"SMP:|\r\n",
				msgType,
				x->x_searchTag);
		return sendto(x->x_write_multicastSocket, message, sizeof(message), 0,
				(struct sockaddr *) x->x_write_dest_addr, x->x_write_addrlen);
	}
	else if( strcasestr(msgType, __testREP)) {
		sprintf(message,	"SABy-%s\r\n"
				"APP:%s\r\n"
				"NID:%s\r\n"
				"TAG:%s\r\n"
				"SMP:%d|%d\r\n",
				msgType,
				x->x_personalData.appField,
				x->x_personalData.name,
				x->x_personalData.tags,
				(int)x->x_personalData.SR,(int)x->x_personalData.framesize);
		int issuer_fd = myClientSocket(x->x_peers[cp].addr,REPLY_TCP_PORT);
		if( issuer_fd == -1 ) {
			error("could not send REP");
			return -1;
		}
		int cnt = sockPuts(issuer_fd,message);
		if( cnt < 0 ) {
			error("Could not send REP message to SRC issuer");
			return -1;
		}
	}
	// ELSE: 
	return -2; // wrong msgType
}










/* --- UTILIY: WAITS FOR A SINGLE TCP CONNECTION ON DESIRED PORT --- */
/* --- inspired by serverSocket from libmsock but handles 1 connection only (no forks) --- */
static int myServerSocket(short port) {

	int sock_fd;
	int dummy = 1;

	/* init address */
	struct sockaddr_in address;
	memset((char *) &address,0,sizeof(address));
	address.sin_family=AF_INET;
	address.sin_port=htons(port);
	address.sin_addr.s_addr=htonl(INADDR_ANY);

	/* get socket */
	sock_fd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);

	/* set REUSE option */
	if( setsockopt(sock_fd,SOL_SOCKET,SO_REUSEADDR,(char *) &dummy, sizeof(int)) < 0) {
		error("netfind: setsockopt fail");
   		return -1;
	}

	/* bind */
	while(bind(sock_fd, (struct sockaddr *) &address, sizeof(address)) < 0) {
		post("netfind: myServerSocket: socket bind fail on port %d, with error %d, will retry in %d secs\n",(short)port,(int)errno,SOCKET_RETRY_INTERVAL);
		sleep(SOCKET_RETRY_INTERVAL); // wait 5 seconds and retry
	}
	return sock_fd;
}

/* --- UTILITY: CONNECT TO ANOTHER TCP ENDPOINT ON A CERTAIN PORT AND ADDRESS --- */
/* --- inspired by clientSocket from libmsock but handles 1 connection only (no forks) --- */
static int myClientSocket(struct in_addr addr, short port) {

	int sock_fd;
	struct sockaddr_in address;
	int cnctd = 0;

	/* init address */
	memset((char *) &address,0,sizeof(address));
	address.sin_family=AF_INET;
	address.sin_port=htons(port);
	address.sin_addr.s_addr=addr.s_addr;

	/* get socket */
	sock_fd = socket(AF_INET, SOCK_STREAM , IPPROTO_TCP);

	cnctd = connect(sock_fd,(struct sockaddr *) &address,sizeof(address));
	if( cnctd < 0 ) {
		perror("CONNECT");
		error("netfind: myClientSocket: ERROR connecting");
		return -1;
	}

	return sock_fd;

}


/* --- UTILITY: STRIP CHARS FROM A STRING */
char *strip_chars(const char *string, const char *chars)
{
	char * newstr = malloc(strlen(string) + 1);
	int counter = 0;

	for ( ; *string; string++) {
		if (!strchr(chars, *string)) {
			newstr[ counter ] = *string;
			++ counter;
		}
	}

	newstr[counter] = 0;
	return newstr;
}



/* --- PARSE UTILITY --- */
int parseAnnounceMsg(t_netfind* x, int cp, char* announceMsg) {

	int i;

	char* tmpptr[MSGLINES+1];

	tmpptr[0] = strcasestr(announceMsg, "APP:");	// point to the APP field content
	tmpptr[1] = strcasestr(announceMsg, "NID:");	// point to the NID field content
	tmpptr[2] = strcasestr(announceMsg, "TAG:");	// point to the TAG field content
	tmpptr[3] = strcasestr(announceMsg, "SMP:");	// point to the SMP field content
	tmpptr[4] = strchr(tmpptr[3],(int)'|');		// point to the framesize in SMP field
	tmpptr[5] = strchrnul(announceMsg, (int)'\0');	// point to EOS

	for(i = 0; i < MSGLINES-1; i++) { // look for troubles or shift the pointers to the actual content.
		if( tmpptr[i] == NULL ) {
			return -1; 
		}
		else {
			tmpptr[i] +=4;
		}
	}
	tmpptr[4] += 1;


	pthread_mutex_lock(&x->x_mutex);
	strncpy(x->x_peers[cp].appField,tmpptr[0],strchr(tmpptr[0],(int)'\r')-tmpptr[0]);
	strncpy(x->x_peers[cp].name,tmpptr[1],strchr(tmpptr[1],(int)'\r')-tmpptr[1]);
	strncpy(x->x_peers[cp].tags,tmpptr[2],strchr(tmpptr[2],(int)'\r')-tmpptr[2]);
	x->x_peers[cp].SR = (t_float)strtof(tmpptr[3], NULL);
	x->x_peers[cp].framesize = atoi(tmpptr[4]);
	pthread_mutex_unlock(&x->x_mutex);
	return 0;
}

/* --- CHECK IF THE PEER cp IS NEW OR IS ME OR IS ALREADY A CONNECTED PEER --- */
int isNewPeer(t_netfind* x, int cp) {
	int i;

	if( strncmp(x->x_peers[cp].name,x->x_personalData.name,MAX_NAME_LEN) == 0 ) {
		return -1; // own message
	}
	for(i = 0; i < x->x_connecState; i++) {
		if ( strncmp(x->x_peers[cp].name,x->x_peers[i].name,MAX_NAME_LEN) == 0 ) {
			return -2; // already connected peer
		}
	}
	return 1;
}

/* --- generate unique ID --- */
void UUID_gen(char *s, const int len) {
    static const char alphanum[] =
        "0123456789"
        "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz";
    int i ;
    srand(time(NULL));
    for (i = 0; i < len; i++) {
        s[i] = alphanum[rand() % (sizeof(alphanum) - 1)];
    }

    s[len] = 0;
}
