/* ------------------------ netreceivev3~ --------------------------------------- */
/*                                                                              */
/* Tilde object to receive uncompressed audio data from netsendv3~.               */
/* Written by Olaf Matthes <olaf.matthes@gmx.de>.                               *
 * Edited by Leonardo Gabrielli <l.gabrielli@univpm.it>                               */
/* Based on streamin~ by Guenter Geiger.                                        */
/*                                                                              */
/* This program is free software; you can redistribute it and/or                */
/* modify it under the terms of the GNU General Public License                  */
/* as published by the Free Software Foundation; either version 3               */
/* of the License, or (at your option) any later version.                       */
/*										*/
/* Copyright (C) 2013 L.Gabrielli						*/
/*                                                                              */
/* See file LICENSE for further informations on licensing terms.                */
/*                                                                              */
/* This program is distributed in the hope that it will be useful,              */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of               */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                */
/* GNU General Public License for more details.                                 */
/*                                                                              */
/* You should have received a copy of the GNU General Public License            */
/* along with this program; If not, see <http://www.gnu.org/licenses/> 		*/
/*                                                                              */
/* Based on PureData by Miller Puckette and others.                             */
/*                                                                              */
/* ---------------------------------------------------------------------------- */

/* DIFF from the original from Olaf matthes *
 * At creation it does not do anything. Awaits for a port number and then binds to it.
 * The port number can arrive from a multicast object.
 * The only creation argument is the number of channels
 * Reduced buffers to cope with latency
 * TCP is not allowed, it increases latency
 * only PD is allowed, will not be tested with MAX
 * FAAC encoding not allowed, it increases latency 
 */


#include "m_pd.h"

#include "netsendv3~.h"


#include <sys/types.h>
#include <string.h>
#ifdef UNIX
#include <sys/socket.h>
#include <errno.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#define SOCKET_ERROR -1
#else
#include <winsock.h>
#endif

#ifndef SOL_IP
#define SOL_IP IPPROTO_IP
#endif

#ifdef NT
#pragma warning( disable : 4244 )
#pragma warning( disable : 4305 )
#endif

#define DEFAULT_AUDIO_BUFFER_FRAMES 4	/* a small circ. buffer for 16 frames */
#define DEFAULT_AVERAGE_NUMBER 10	/* number of values we store for average history */
#define DEFAULT_NETWORK_POLLTIME 1	/* interval in ms for polling for input data (Max/MSP only) */
#define DEFAULT_QUEUE_LENGTH 3		/* min. number of buffers that can be used reliably on your hardware */

#ifdef UNIX
#define CLOSESOCKET(fd) close(fd)
#endif
#ifdef _WINDOWS
#define CLOSESOCKET(fd) closesocket(fd)
#endif

/* these would require to include some headers that are different
   between pd 0.36 and later, so it's easier to do it like this! */
EXTERN void sys_rmpollfn(int fd);
EXTERN void sys_addpollfn(int fd, void* fn, void *ptr);

static int netreceivev3_tilde_sockerror(char *s)
{
#ifdef NT
    int err = WSAGetLastError();
    if (err == 10054) return 1;
    else if (err == 10040) post("netsendv3~: %s: message too long (%d)", s, err);
    else if (err == 10053) post("netsendv3~: %s: software caused connection abort (%d)", s, err);
    else if (err == 10055) post("netsendv3~: %s: no buffer space available (%d)", s, err);
    else if (err == 10060) post("netsendv3~: %s: connection timed out (%d)", s, err);
    else if (err == 10061) post("netsendv3~: %s: connection refused (%d)", s, err);
    else post("netreceivev3~: %s: %s (%d)", s, strerror(err), err);
#else
    int err = errno;
    post("netreceivev3~: %s: %s (%d)", s, strerror(err), err);
#endif
#ifdef NT
	if (err == WSAEWOULDBLOCK)
#endif
#ifdef UNIX
	if (err == EAGAIN)
#endif
	{
		return 1;	/* recoverable error */
	}
	return 0;	/* indicate non-recoverable error */
}


static int netreceivev3_tilde_setsocketoptions(int sockfd)
{ 
	int sockopt = 1;
	if (setsockopt(sockfd, SOL_IP, TCP_NODELAY, (const char*)&sockopt, sizeof(int)) < 0)
		post("setsockopt NODELAY failed");

	sockopt = 1;    
	if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, (const char*)&sockopt, sizeof(int)) < 0)
		post("netreceivev3~: setsockopt REUSEADDR failed");
	return 0;
}



/* ------------------------ netreceivev3~ ----------------------------- */
static t_class *netreceivev3_tilde_class;
static t_symbol *ps_format, *ps_channels, *ps_framesize, *ps_overflow, *ps_underflow,
                *ps_queuesize, *ps_average, *ps_sf_float, *ps_sf_16bit, *ps_sf_8bit, 
                *ps_sf_mp3, *ps_sf_aac, *ps_sf_unknown, *ps_bitrate, *ps_hostname, *ps_nothing;


typedef struct _netreceivev3_tilde
{
	t_object x_obj;
	t_outlet *x_outlet1;
	t_outlet *x_outlet2;
	t_outlet *x_outlet_loss;

	int x_socket;
	int x_connectsocket;
	int x_nconnections;
	int x_ndrops;
	int x_tcp;
	t_symbol *x_hostname;

	/* buffering */
	int x_framein;
	int x_frameout;
	t_frame x_frames[DEFAULT_AUDIO_BUFFER_FRAMES];
	int x_maxframes;
	long x_framecount;
	int x_blocksize;
	int x_blocksperrecv;
	int x_blockssincerecv;

	int x_nbytes;
	int x_counter;
	int x_average[DEFAULT_AVERAGE_NUMBER];
	int x_averagecur;
	int x_underflow;
	int x_overflow;

	long x_samplerate;
	int x_noutlets;
	int x_vecsize;
	t_int **x_myvec;            /* vector we pass on to the DSP routine */
} t_netreceivev3_tilde;



/* prototypes (as needed) */
static void netreceivev3_tilde_kick(t_netreceivev3_tilde *x);






/* remove all pollfunctions and close socket */
static void netreceivev3_tilde_closesocket(t_netreceivev3_tilde* x)
{
	sys_rmpollfn(x->x_socket);
	outlet_float(x->x_outlet1, 0);

	CLOSESOCKET(x->x_socket);
	x->x_socket = -1;
}



static void netreceivev3_tilde_reset(t_netreceivev3_tilde* x, t_floatarg buffer)
{
	int i;
	x->x_counter = 0;
	x->x_nbytes = 0;
	x->x_framein = 0;
	x->x_frameout = 0;
	x->x_blockssincerecv = 0;
	x->x_blocksperrecv = x->x_blocksize / x->x_vecsize;

	for (i = 0; i < DEFAULT_AVERAGE_NUMBER; i++)
		x->x_average[i] = x->x_maxframes;
	x->x_averagecur = 0;

	if (buffer == 0.0)	/* set default */
		x->x_maxframes = DEFAULT_QUEUE_LENGTH;
	else
	{
		buffer = (float)CLIP((float)buffer, 0., 1.);
		x->x_maxframes = (int)(DEFAULT_AUDIO_BUFFER_FRAMES * buffer);
		x->x_maxframes = CLIP(x->x_maxframes, 1, DEFAULT_AUDIO_BUFFER_FRAMES - 1);
		post("netreceivev3~: set buffer to %g (%d frames)", buffer, x->x_maxframes);
	}
	x->x_underflow = 0;
	x->x_overflow = 0;
}


static void netreceivev3_tilde_datapoll(t_netreceivev3_tilde *x)
{

	int ret;
	int n;

		
	/* UDP */
	n = x->x_nbytes;
	if (x->x_nbytes == 0)	/* we ate all the samples and need a new header tag */
	{
		/* receive header tag */
		ret = recv(x->x_socket, (char*)&x->x_frames[x->x_framein].tag, sizeof(t_tag), 0);
		if (ret <= 0)	/* error */
		{
			if (netreceivev3_tilde_sockerror("recv tag"))
				goto bail;
			netreceivev3_tilde_reset(x, 0);
			x->x_counter = 0;
			return;
		}
		else if (ret != sizeof(t_tag))
		{
			/* incomplete header tag: return and try again later */
			/* in the hope that more data will be available */
			error("netreceivev3~: got incomplete header tag");
			return;
		}
		/* adjust byte order if neccessarry */
		if (x->x_frames[x->x_framein].tag.version != SF_BYTE_NATIVE)
		{
			x->x_frames[x->x_framein].tag.count = netsendv3_long(x->x_frames[x->x_framein].tag.count);
			x->x_frames[x->x_framein].tag.framesize = netsendv3_long(x->x_frames[x->x_framein].tag.framesize);
		}
		/* get info from header tag */
		if (x->x_frames[x->x_framein].tag.channels > x->x_noutlets)
		{
			error("netreceivev3~: incoming stream has too many channels (%d)", x->x_frames[x->x_framein].tag.channels);
			x->x_counter = 0;
			return;
		}
		x->x_nbytes = n = x->x_frames[x->x_framein].tag.framesize;
	}
	else	/* we already have header tag or some data and need more */
	{
		ret = recv(x->x_socket, (char*)x->x_frames[x->x_framein].data + x->x_frames[x->x_framein].tag.framesize - n, n, 0);
		if (ret > 0)
		{
			n -= ret;
		}
		else if (ret < 0)	/* error */
		{
			if (netreceivev3_tilde_sockerror("recv data"))
				goto bail;
			netreceivev3_tilde_reset(x, 0);
			x->x_counter = 0;
			return;
		}
		x->x_nbytes = n;
		if (n == 0)			/* a complete packet is received */
		{
			if (x->x_frames[x->x_framein].tag.format == SF_AAC)
			{
				error("netreceivev3~: don't know how to decode AAC format");
				return;
			}
			x->x_counter++;
			x->x_framein++;
			x->x_framein %= DEFAULT_AUDIO_BUFFER_FRAMES;

			/* check for buffer overflow */
			if (x->x_framein == x->x_frameout)
			{
				x->x_overflow++;
			}
		}
	}
		
	
bail:
	;
}


static void netreceivev3_tilde_connectpoll(t_netreceivev3_tilde *x)
{
	
	int sockaddrl = (int)sizeof(struct sockaddr);
	struct sockaddr_in incomer_address;
	int fd = accept(x->x_connectsocket, (struct sockaddr*)&incomer_address, &sockaddrl);
	if (fd < 0) 
	{
		post("netreceivev3~: accept failed");
		return;
	}
#ifdef O_NONBLOCK
	fcntl(fd, F_SETFL, O_NONBLOCK);
#endif
	if (x->x_socket != -1)
	{
		post("netreceivev3~: new connection");
		netreceivev3_tilde_closesocket(x);
	}

	netreceivev3_tilde_reset(x, 0);
	x->x_socket = fd;
	x->x_nbytes = 0;
	x->x_hostname = gensym(inet_ntoa(incomer_address.sin_addr));

	sys_addpollfn(fd, netreceivev3_tilde_datapoll, x);
	outlet_float(x->x_outlet1, 1);

}


static int netreceivev3_tilde_createsocket(t_netreceivev3_tilde* x, int portno)
{
	struct sockaddr_in server;
	int sockfd;
	int tcp = x->x_tcp;

	/* create a socket */
	if (!tcp) {
		sockfd = socket(AF_INET, SOCK_DGRAM, 0);
	}
	else {
		sockfd = socket(AF_INET, SOCK_STREAM, 0);
		post("netreceivev3~: creating a TCP socket! (UDP recommended: put 3rd argument = 1)");
	}

	if (sockfd < 0)
	{
		netreceivev3_tilde_sockerror("socket");
		return 0;
	}
	server.sin_family = AF_INET;
	server.sin_addr.s_addr = INADDR_ANY;

	/* assign server port number */

	server.sin_port = htons((u_short)portno);
	post("netreceivev3~: listening to port number %d", portno);

	netreceivev3_tilde_setsocketoptions(sockfd);

	/* name the socket */
	if (bind(sockfd, (struct sockaddr *)&server, sizeof(server)) < 0)
	{
		netreceivev3_tilde_sockerror("bind");
		CLOSESOCKET(sockfd);
		return 0;
	}


	if (!tcp)
	{
		x->x_socket = sockfd;
		x->x_nbytes = 0;
		sys_addpollfn(sockfd, netreceivev3_tilde_datapoll, x); // questa funzione viene aggiunta alla creazione del socket... è un problema se il socket non viene creato all'avvio?
	}
	else
	{
		if (listen(sockfd, 5) < 0)
		{
			netreceivev3_tilde_sockerror("listen");
			CLOSESOCKET(sockfd);
			return 0;
		}
		else
		{
			x->x_connectsocket = sockfd;
			/* start polling for connection requests */
			sys_addpollfn(sockfd, netreceivev3_tilde_connectpoll, x);
		}
	}
	return 1;
}



/* kick connected client */
static void netreceivev3_tilde_kick(t_netreceivev3_tilde *x)
{
	if (x->x_tcp)
	{
		if (x->x_socket != -1)
		{
			shutdown(x->x_socket, 1);
			netreceivev3_tilde_closesocket(x);
			post("netreceivev3~: kicked client!");
		}
		else error("netreceivev3~: no client to kick");
	}
	else error("netreceivev3~: kicking clients in UDP mode not possible");
}


#define QUEUESIZE (int)((x->x_framein + DEFAULT_AUDIO_BUFFER_FRAMES - x->x_frameout) % DEFAULT_AUDIO_BUFFER_FRAMES)
#define BLOCKOFFSET (x->x_blockssincerecv * x->x_vecsize * x->x_frames[x->x_frameout].tag.channels)

static t_int *netreceivev3_tilde_perform(t_int *w)
{
	t_netreceivev3_tilde *x = (t_netreceivev3_tilde*) (w[1]);
	int n = (int)(w[2]);
	t_float *out[DEFAULT_AUDIO_CHANNELS];
	const int offset = 3;
	const int channels = x->x_frames[x->x_frameout].tag.channels;
	int i = 0;

	for (i = 0; i < x->x_noutlets; i++)
	{
		out[i] = (t_float *)(w[offset + i]);
	}

	if (n != x->x_vecsize)
	{
		x->x_vecsize = n;
		x->x_blocksperrecv = x->x_blocksize / x->x_vecsize;
		x->x_blockssincerecv = 0;
	}

	/* check whether there is enough data in buffer */
	if (x->x_counter < x->x_maxframes)
	{
		goto bail;
	}

	/* check for buffer underflow */
	if (x->x_framein == x->x_frameout)
	{
		x->x_underflow++;
		goto bail;
	}


	/* queue balancing */
	x->x_average[x->x_averagecur] = QUEUESIZE;
	if (++x->x_averagecur >= DEFAULT_AVERAGE_NUMBER)
		x->x_averagecur = 0;

	switch (x->x_frames[x->x_frameout].tag.format)
	{
		case SF_FLOAT:
		{
			t_float* buf = (t_float *)x->x_frames[x->x_frameout].data + BLOCKOFFSET;

			if (x->x_frames[x->x_frameout].tag.version == SF_BYTE_NATIVE)
			{
				while (n-- > 0)
				{
					for (i = 0; i < channels; i++)
					{     
						*out[i]++ = *buf++;
					}
					for (i = channels; i < x->x_noutlets; i++)
					{     
						*out[i]++ = 0.;
					}
				}
			}
			else	/* swap bytes */
			{
				while (n-- > 0)
				{
					for (i = 0; i < channels; i++)
					{     
						*out[i]++ = netsendv3_float(*buf++);
					}
					for (i = channels; i < x->x_noutlets; i++)
					{     
						*out[i]++ = 0.;
					}
				}
			}
			break;
		}
		case SF_16BIT:     
		{
			short* buf = (short *)x->x_frames[x->x_frameout].data + BLOCKOFFSET;

			if (x->x_frames[x->x_frameout].tag.version == SF_BYTE_NATIVE)
			{
				while (n-- > 0)
				{
					for (i = 0; i < channels; i++)
					{
						*out[i]++ = (t_float)(*buf++ * 3.051850e-05);
					}
					for (i = channels; i < x->x_noutlets; i++)
					{
						*out[i]++ = 0.;
					}
				}
			}
			else /* swap bytes */
			{
				while (n-- > 0)
				{
					for (i = 0; i < channels; i++)
					{
						*out[i]++ = (t_float)(netsendv3_short(*buf++) * 3.051850e-05);
					}
					for (i = channels; i < x->x_noutlets; i++)
					{
						*out[i]++ = 0.;
					}
				}
			}
			break;
		}
		case SF_8BIT:     
		{
			unsigned char* buf = (char *)x->x_frames[x->x_frameout].data + BLOCKOFFSET;

			while (n-- > 0)
			{
				for (i = 0; i < channels; i++)
				{ 
					*out[i]++ = (t_float)((0.0078125 * (*buf++)) - 1.0);
				}
				for (i = channels; i < x->x_noutlets; i++)
				{
					*out[i]++ = 0.;
				}
			}
			break;
		}
		case SF_MP3:     
		{
			post("netreceivev3~: mp3 format not supported");
			if (x->x_tcp)
				netreceivev3_tilde_kick(x);
			break;
		}
		case SF_AAC:     
		{
			post("netreceivev3~: aac format not supported");
			if (x->x_tcp)
				netreceivev3_tilde_kick(x);
			break;
		}
		default:
			post("netreceivev3~: unknown format (%d)",x->x_frames[x->x_frameout].tag.format);
			if (x->x_tcp)
				netreceivev3_tilde_kick(x);
			break;
	}

	if (!(x->x_blockssincerecv < x->x_blocksperrecv - 1))
	{
		x->x_blockssincerecv = 0;
		x->x_frameout++;
		x->x_frameout %= DEFAULT_AUDIO_BUFFER_FRAMES;
	}
	else
	{
		x->x_blockssincerecv++;
	}

	return (w + offset + x->x_noutlets);

bail:
	/* set output to zero */
	while (n-- > 0)
	{
		for (i = 0; i < x->x_noutlets; i++)
		{
			*(out[i]++) = 0.;
		}
	}
	return (w + offset + x->x_noutlets);
}



static void netreceivev3_tilde_dsp(t_netreceivev3_tilde *x, t_signal **sp)
{
	int i;

	x->x_myvec[0] = (t_int*)x;
	x->x_myvec[1] = (t_int*)sp[0]->s_n;

	x->x_samplerate = (long)sp[0]->s_sr;

	if (DEFAULT_AUDIO_BUFFER_SIZE % sp[0]->s_n)
	{
		error("netsendv3~: signal vector size too large (needs to be even divisor of %d)", DEFAULT_AUDIO_BUFFER_SIZE);
	}
	else
	{
		for (i = 0; i < x->x_noutlets; i++)
		{
			x->x_myvec[2 + i] = (t_int*)sp[i + 1]->s_vec; // i+1 perché sp[0] è il segnale dell'inlet (invisibile?)
		}
		dsp_addv(netreceivev3_tilde_perform, x->x_noutlets + 2, (t_int*)x->x_myvec);
	}
}


/* send stream info when banged */
static void netreceivev3_tilde_bang(t_netreceivev3_tilde *x)
{
	t_atom list[2];
	t_symbol *sf_format;
	t_float bitrate;
	int i, avg = 0;
	for (i = 0; i < DEFAULT_AVERAGE_NUMBER; i++)
		avg += x->x_average[i];

	bitrate = (t_float)((SF_SIZEOF(x->x_frames[x->x_frameout].tag.format) * x->x_samplerate * 8 * x->x_frames[x->x_frameout].tag.channels) / 1000.);

	switch (x->x_frames[x->x_frameout].tag.format)
	{
		case SF_FLOAT:
		{
			sf_format = ps_sf_float;
			break;
		}
		case SF_16BIT:
		{
			sf_format = ps_sf_16bit;
			break;
		}
		case SF_8BIT:
		{
			sf_format = ps_sf_8bit;
			break;
		}
		case SF_MP3:
		{
			sf_format = ps_sf_mp3;
			break;
		}
		case SF_AAC:
		{
			sf_format = ps_sf_aac;
			break;
		}
		default:
		{
			sf_format = ps_sf_unknown;
			break;
		}
	}


	/* --- stream information (t_tag) --- */
	/* audio format */
	SETSYMBOL(list, (t_symbol *)sf_format);
	outlet_anything(x->x_outlet2, ps_format, 1, list);

	/* channels */
	SETFLOAT(list, (t_float)x->x_frames[x->x_frameout].tag.channels);
	outlet_anything(x->x_outlet2, ps_channels, 1, list);

	/* framesize */
	SETFLOAT(list, (t_float)x->x_frames[x->x_frameout].tag.framesize);
	outlet_anything(x->x_outlet2, ps_framesize, 1, list);

	/* bitrate */
	SETFLOAT(list, (t_float)bitrate);
	outlet_anything(x->x_outlet2, ps_bitrate, 1, list);

	/* --- internal info (buffer and network) --- */
	/* overflow */
	SETFLOAT(list, (t_float)x->x_overflow);
	outlet_anything(x->x_outlet2, ps_overflow, 1, list);

	/* underflow */
	SETFLOAT(list, (t_float)x->x_underflow);
	outlet_anything(x->x_outlet2, ps_underflow, 1, list);

	/* queuesize */
	SETFLOAT(list, (t_float)QUEUESIZE);
	outlet_anything(x->x_outlet2, ps_queuesize, 1, list);

	/* average queuesize */
	SETFLOAT(list, (t_float)((t_float)avg / (t_float)DEFAULT_AVERAGE_NUMBER));
	outlet_anything(x->x_outlet2, ps_average, 1, list);

	if (x->x_tcp)
	{
		/* IP address */
		SETSYMBOL(list, (t_symbol *)x->x_hostname);
		outlet_anything(x->x_outlet2, ps_hostname, 1, list);
	}
}



static void netreceivev3_tilde_print(t_netreceivev3_tilde* x)
{
	int i, avg = 0;
	for (i = 0; i < DEFAULT_AVERAGE_NUMBER; i++)
		avg += x->x_average[i];
	post("netreceivev3~: last size = %d, avg size = %g, %d underflows, %d overflows", QUEUESIZE, (float)((float)avg / (float)DEFAULT_AVERAGE_NUMBER), x->x_underflow, x->x_overflow);
	post("netreceivev3~: channels = %d, framesize = %d, packets = %d", x->x_frames[x->x_framein].tag.channels, x->x_frames[x->x_framein].tag.framesize, x->x_counter);
}




static void *netreceivev3_tilde_new(t_floatarg outlets)
{
	t_netreceivev3_tilde *x;
	int i;

	x = (t_netreceivev3_tilde *)pd_new(netreceivev3_tilde_class);
    if (x)
    { 
        for (i = sizeof(t_object); i < (int)sizeof(t_netreceivev3_tilde); i++)
                ((char *)x)[i] = 0; 
	}

	x->x_noutlets = CLIP((int)outlets, 1, DEFAULT_AUDIO_CHANNELS);
	for (i = 0; i < x->x_noutlets; i++)
		outlet_new(&x->x_obj, &s_signal);
	//if (!prot) // se il terzo argomento == 0 TCP (quindi apro un outlet controllo x comunicare lo stato della connessione ndLeo)
		x->x_outlet1 = outlet_new(&x->x_obj, &s_anything);	/* outlet for connection state (TCP/IP) */
	x->x_outlet2 = outlet_new(&x->x_obj, &s_anything);

	x->x_myvec = (t_int **)t_getbytes(sizeof(t_int *) * (x->x_noutlets + 3));
	if (!x->x_myvec)
	{
		error("netreceivev3~: out of memory");
		return NULL;
	}

	x->x_connectsocket = -1;
	x->x_socket = -1;
	x->x_tcp = 0; // UDP is mandatory
	x->x_nconnections = 0;
	x->x_ndrops = 0;
	x->x_underflow = 0;
	x->x_overflow = 0;
	x->x_hostname = ps_nothing;

	for (i = 0; i < DEFAULT_AUDIO_BUFFER_FRAMES; i++)
	{
		x->x_frames[i].data = (char *)t_getbytes(DEFAULT_AUDIO_BUFFER_SIZE * x->x_noutlets * sizeof(t_float));
	}
	x->x_framein = 0;
	x->x_frameout = 0;
	x->x_maxframes = DEFAULT_QUEUE_LENGTH;
	x->x_vecsize = 64;	/* we'll update this later */
	x->x_blocksize = DEFAULT_AUDIO_BUFFER_SIZE;	/* LATER make this dynamic */
	x->x_blockssincerecv = 0;
	x->x_blocksperrecv = x->x_blocksize / x->x_vecsize;

	/* here the original netreceive would create a socket, we wait instead for info from netfind or a port message */

	return (x);
}


static void netreceivev3_tilde_port(t_netreceivev3_tilde* x, t_floatarg fportno) {

	if (!netreceivev3_tilde_createsocket(x, (int)fportno))
	{
		error("netreceivev3~: failed to create listening socket");
		return (NULL);
	}

}

static void netreceivev3_tilde_free(t_netreceivev3_tilde *x)
{
	int i;

	if (x->x_connectsocket != -1)
	{
		sys_rmpollfn(x->x_connectsocket);
		CLOSESOCKET(x->x_connectsocket);
	}
	if (x->x_socket != -1)
	{
		sys_rmpollfn(x->x_socket);
		CLOSESOCKET(x->x_socket);
	}

	/* free memory */
	t_freebytes(x->x_myvec, sizeof(t_int *) * (x->x_noutlets + 3));
	for (i = 0; i < DEFAULT_AUDIO_BUFFER_FRAMES; i++)
	{
		 t_freebytes(x->x_frames[i].data, DEFAULT_AUDIO_BUFFER_SIZE * x->x_noutlets * sizeof(t_float));
	}
}




void netreceivev3_tilde_setup(void)
{
	netreceivev3_tilde_class = class_new(gensym("netreceivev3~"),
		(t_newmethod) netreceivev3_tilde_new, (t_method) netreceivev3_tilde_free,
		sizeof(t_netreceivev3_tilde),  0, A_DEFFLOAT, A_NULL);

	class_addmethod(netreceivev3_tilde_class, nullfn, gensym("signal"), 0);
	class_addbang(netreceivev3_tilde_class, (t_method)netreceivev3_tilde_bang);
	class_addmethod(netreceivev3_tilde_class, (t_method)netreceivev3_tilde_dsp, gensym("dsp"), 0);
	class_addmethod(netreceivev3_tilde_class, (t_method)netreceivev3_tilde_print, gensym("print"), 0);
	class_addmethod(netreceivev3_tilde_class, (t_method)netreceivev3_tilde_kick, gensym("kick"), 0);
	class_addmethod(netreceivev3_tilde_class, (t_method)netreceivev3_tilde_reset, gensym("reset"), A_DEFFLOAT, 0);
	class_addmethod(netreceivev3_tilde_class, (t_method)netreceivev3_tilde_port, gensym("port"), A_DEFFLOAT, 0);
	class_sethelpsymbol(netreceivev3_tilde_class, gensym("netsendv3~"));

	ps_format = gensym("format");
	ps_channels = gensym("channels");
	ps_framesize = gensym("framesize");
	ps_bitrate = gensym("bitrate");
	ps_overflow = gensym("overflow");
	ps_underflow = gensym("underflow");
	ps_queuesize = gensym("queuesize");
	ps_average = gensym("average");
	ps_hostname = gensym("ipaddr");
	ps_sf_float = gensym("_float_");
	ps_sf_16bit = gensym("_16bit_");
	ps_sf_8bit = gensym("_8bit_");
	ps_sf_mp3 = gensym("_mp3_");
	ps_sf_aac = gensym("_aac_");
	ps_sf_unknown = gensym("_unknown_");
	ps_nothing = gensym("");
}

