/* ------------------------ nettest~ ------------------------------------------ */
/*                                                                              */
/* Pure Data external to test audio network latency between Puredata machines.  */
/* nettest~ must be instantiated in one machine with argument 1 and on the other*/
/* with argument 2.								*/
/* Written by Leonardo Gabrielli <l.gabrielli@univpm.it>			*/
/* Partly based on netreceive and netsend by Olaf Matthes			*/
/* Source and updates at http://a3lab.dii.univpm/wemust/netfind 		*/
/*										*/
/* Reference paper for the SABy protocol: 					*/
/* L. Gabrielli, S.Squartini, F. Piazza, “Advancements and Performance analysis */
/* on the Wireless Music Studio (WeMUST) framework”, in Proc. AES 134th 	*/
/* Convention, Rome, Italy, 2013						*/
/*										*/
/* Copyright (C) 2013 L.Gabrielli						*/
/*										*/
/* This program is free software; you can redistribute it and/or                */
/* modify it under the terms of the GNU General Public License                  */
/* as published by the Free Software Foundation; either version 3               */
/* of the License, or (at your option) any later version.                       */
/*                                                                              */
/* See file LICENSE for further informations on licensing terms.                */
/*                                                                              */
/* This program is distributed in the hope that it will be useful,              */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of               */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                */
/* GNU General Public License for more details.                                 */
/*                                                                              */
/* You should have received a copy of the GNU General Public License            */
/* along with this program; If not, see <http://www.gnu.org/licenses/> 		*/
/*                                                                              */
/* Based on PureData by Miller Puckette and others.                             */
/*                                                                              */
/* ---------------------------------------------------------------------------- */


#include "m_pd.h"

#include <sys/types.h>
#include "nettest~.h"
#include "float_cast.h"	/* tools for fast conversion from float to int */

#include <sys/types.h>
#include <string.h>
#include <stdlib.h>
#ifdef UNIX
#include <sys/socket.h>
#include <errno.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/time.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <pthread.h>
#define SOCKET_ERROR -1
#endif
#ifdef _WINDOWS
#include <winsock.h>
#include "pthread.h"
#endif

#ifdef _WINDOWS
#pragma warning( disable : 4244 )
#pragma warning( disable : 4305 )
#endif

#ifdef MSG_NOSIGNAL
#define SEND_FLAGS /*MSG_DONTWAIT|*/MSG_NOSIGNAL
#else
#define SEND_FLAGS 0
#endif

#ifndef SOL_IP
#define SOL_IP IPPROTO_IP
#endif

/* nr */
#ifdef NT
#pragma warning( disable : 4244 )
#pragma warning( disable : 4305 )
#endif

#define DEFAULT_AUDIO_BUFFER_FRAMES 8		/* a small circ. buffer for 16 frames */
#define DEFAULT_AVERAGE_NUMBER 10		/* number of values we store for average history */
#define DEFAULT_NETWORK_POLLTIME 1		/* interval in ms for polling for input data (Max/MSP only) */
#define DEFAULT_QUEUE_LENGTH 3			/* min. number of buffers that can be used reliably on your hardware */

#ifdef UNIX
#define CLOSESOCKET(fd) close(fd)
#endif
#ifdef _WINDOWS
#define CLOSESOCKET(fd) closesocket(fd)
#endif

#ifdef PD
/* these would require to include some headers that are different
   between pd 0.36 and later, so it's easier to do it like this! */
EXTERN void sys_rmpollfn(int fd);
EXTERN void sys_addpollfn(int fd, void* fn, void *ptr);
#endif

static int receive_sockerror(char *s)
{
#ifdef NT
    int err = WSAGetLastError();
    if (err == 10054) return 1;
    else if (err == 10040) post("nettest~: %s: message too long (%d)", s, err);
    else if (err == 10053) post("nettest~: %s: software caused connection abort (%d)", s, err);
    else if (err == 10055) post("nettest~: %s: no buffer space available (%d)", s, err);
    else if (err == 10060) post("nettest~: %s: connection timed out (%d)", s, err);
    else if (err == 10061) post("nettest~: %s: connection refused (%d)", s, err);
    else post("netreceivev3~: %s: %s (%d)", s, strerror(err), err);
#else
    int err = errno;
    post("netreceivev3~: %s: %s (%d)", s, strerror(err), err);
#endif
#ifdef NT
	if (err == WSAEWOULDBLOCK)
#endif
#ifdef UNIX
	if (err == EAGAIN)
#endif
	{
		return 1;	/* recoverable error */
	}
	return 0;	/* indicate non-recoverable error */
}


static int receive_setsocketoptions(int sockfd)
{
	int sockopt = 1;
	if (setsockopt(sockfd, SOL_IP, TCP_NODELAY, (const char*)&sockopt, sizeof(int)) < 0)
		post("setsockopt NODELAY failed");

	sockopt = 1;
	if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, (const char*)&sockopt, sizeof(int)) < 0)
		post("netreceivev3~: setsockopt REUSEADDR failed");
	return 0;
}


/* --- nettest_tilde CLASS --- */
static t_class *nettest_tilde_class;

static t_symbol *ps_nothing, *ps_localhost;
static t_symbol *ps_format, *ps_channels, *ps_framesize, *ps_overflow, *ps_underflow;
static t_symbol *ps_queuesize, *ps_average, *ps_sf_float, *ps_sf_16bit, *ps_sf_8bit;
static t_symbol *ps_sf_mp3, *ps_sf_aac, *ps_sf_unknown, *ps_bitrate, *ps_hostname;

/* nr */
static t_symbol *ps_format, *ps_channels, *ps_framesize, *ps_overflow, *ps_underflow,
                *ps_queuesize, *ps_average, *ps_sf_float, *ps_sf_16bit, *ps_sf_8bit,
                *ps_sf_mp3, *ps_sf_aac, *ps_sf_unknown, *ps_bitrate, *ps_hostname, *ps_nothing;

typedef struct _nettest_tilde
{
	t_object x_obj;
	t_outlet *x_outlet;
	t_outlet *x_outlet2;
	t_clock *x_clock;



	int x_fd;
	int x_protocol;
	t_tag x_tag;
	t_symbol* x_hostname;
	int x_portno;
	int x_connectstate;
	char *x_cbuf;
	int x_cbufsize;
	int x_blocksize;
	int x_blockspersend;
	int x_blockssincesend;

	long x_samplerate;          /* samplerate we're running at */
	int x_vecsize;              /* current DSP signal vector size */
	int x_ninlets;              /* number of inlets */
	int x_channels;             /* number of channels we want to stream */
	int x_format;               /* format of streamed audio data */
	int x_bitrate;              /* specifies bitrate for compressed formats */
	int x_count;                /* total number of audio frames */ //ToDo: should it not be LONG? x_tag.count is LONG
	t_int **x_myvec;            /* vector we pass on in the DSP routine */



	pthread_mutex_t   x_mutex;
	pthread_cond_t    x_requestcondition;
	pthread_cond_t    x_answercondition;
	pthread_t         x_childthread;

	/* mutuated from netreceive */
	int x_socket;
	int x_connectsocket;
	int x_nconnections;
	int x_tcp;
	t_symbol *x_nr_hostname;

	t_outlet *x_outlet3;
	t_outlet *x_outlet4;
	t_outlet *x_outlet_loss;
	t_outlet *x_outlet_lat;
	
	/* buffering */
	int x_framein;
	int x_frameout;
	t_frame x_frames[DEFAULT_AUDIO_BUFFER_FRAMES];
	int x_maxframes;
	long x_framecount;
	int x_nr_blocksize;
	int x_blocksperrecv;
	int x_blockssincerecv;

	int x_nbytes;
	int x_counter;
	int x_average[DEFAULT_AVERAGE_NUMBER];
	int x_averagecur;
	int x_underflow;
	int x_overflow;

	long x_nr_samplerate;
	int x_noutlets;
	int x_nr_vecsize;
	t_int **x_nr_myvec;	/* vector we pass on to the DSP routine */

	int x_mode;		/* 1 = sends and outputs latency , 2 = sends back everything to 1 */
	t_float x_float;

	t_float x_latency;	/* latency in ms */

	t_float* x_tmpbuf_start[DEFAULT_AUDIO_CHANNELS];

} t_nettest_tilde;



/* --- PROTOTYPES --- */
static int nettest_tilde_sockerror(char *s);
static void nettest_tilde_closesocket(int fd);
static void nettest_tilde_notify(t_nettest_tilde *x);
static void nettest_tilde_disconnect(t_nettest_tilde *x);
static void *nettest_tilde_doconnect(void *zz);
static void nettest_tilde_connect(t_nettest_tilde *x, t_symbol *host, t_floatarg fportno);
static t_int *nettest_tilde_perform(t_int *w);
static void nettest_tilde_dsp(t_nettest_tilde *x, t_signal **sp);
static void nettest_tilde_channels(t_nettest_tilde *x, t_floatarg channels);
static void nettest_tilde_format(t_nettest_tilde *x, t_symbol* form, t_floatarg bitrate);
static void nettest_tilde_host(t_nettest_tilde *x, t_symbol* host);
static void nettest_tilde_float(t_nettest_tilde* x, t_floatarg arg);
static void nettest_tilde_bang(t_nettest_tilde *x);
static void nettest_tilde_free(t_nettest_tilde* x);

/* nr */
static void receive_kick(t_nettest_tilde *x);
static void receive_closesocket(t_nettest_tilde* x);
static void receive_reset(t_nettest_tilde* x, t_floatarg buffer);
static void receive_datapoll(t_nettest_tilde *x);
static void nettest_tilde_print(t_nettest_tilde* x);
static void receive_port(t_nettest_tilde* x, t_floatarg fportno);
void nettest_latency_calc(t_nettest_tilde*, int, int);
void nettest_latency_output(t_nettest_tilde*);

/* --- nettest_TILDE NEW METHOD --- */
static void *nettest_tilde_new(t_floatarg mode)
{
	int i;
	int inlets = 2;	/* default test: 2 channels */
	int outlets = inlets;

	t_nettest_tilde *x = (t_nettest_tilde *)pd_new(nettest_tilde_class);
	if (x)
	{
		for (i = sizeof(t_object); i < (int)sizeof(t_nettest_tilde); i++)
			((char *)x)[i] = 0;
	}

	x->x_ninlets = CLIP((int)inlets, 1, DEFAULT_AUDIO_CHANNELS);
	x->x_noutlets = x->x_ninlets;

	/* additional inlets if channels > 1 */
	for (i = 1; i < x->x_ninlets; i++) 
		inlet_new(&x->x_obj, &x->x_obj.ob_pd, &s_signal, &s_signal);
	
	/* outlets */
	for (i = 0; i < x->x_noutlets; i++)
		outlet_new(&x->x_obj, &s_signal);


	x->x_outlet = outlet_new(&x->x_obj, &s_float);
	x->x_outlet2 = outlet_new(&x->x_obj, &s_list);
	x->x_outlet3 = outlet_new(&x->x_obj, &s_anything);
	x->x_outlet4 = outlet_new(&x->x_obj, &s_anything);
	x->x_outlet_lat = outlet_new(&x->x_obj, &s_anything);
	x->x_clock = clock_new(x, (t_method)nettest_tilde_notify);

	x->x_mode = mode;

	x->x_myvec = (t_int **)t_getbytes(sizeof(t_int *) * (x->x_ninlets + x->x_noutlets + 3));
	if (!x->x_myvec)
	{
		error("nettest~: out of memory");
		return NULL;
	}

	x->x_connectsocket = -1;
	x->x_socket = -1;
	x->x_tcp = 0; /* udo only */
	x->x_nconnections = 0;
	x->x_underflow = 0;
	x->x_overflow = 0;
	x->x_nr_hostname = ps_nothing;

	for (i = 0; i < DEFAULT_AUDIO_BUFFER_FRAMES; i++)
	{
		x->x_frames[i].data = (char *)t_getbytes(DEFAULT_AUDIO_BUFFER_SIZE * x->x_noutlets * sizeof(t_float));
	}
	x->x_framein = 0;
	x->x_frameout = 0;
	x->x_maxframes = DEFAULT_QUEUE_LENGTH;
	x->x_nr_vecsize = 64;	/* we'll update this later */
	x->x_nr_blocksize = DEFAULT_AUDIO_BUFFER_SIZE;	/* LATER make this dynamic */
	x->x_blockssincerecv = 0;
	x->x_blocksperrecv = x->x_nr_blocksize / x->x_nr_vecsize;

	pthread_mutex_init(&x->x_mutex, 0);
	pthread_cond_init(&x->x_requestcondition, 0);
	pthread_cond_init(&x->x_answercondition, 0);

	x->x_hostname = ps_localhost;
	x->x_portno = 0;
	x->x_connectstate = 0;
	x->x_childthread = 0;
	x->x_fd = -1;
	x->x_protocol = SOCK_DGRAM;


	x->x_tag.format = x->x_format = SF_FLOAT;
	x->x_tag.channels = x->x_channels = x->x_ninlets;
	x->x_tag.version = SF_BYTE_NATIVE;	/* native endianness */
	x->x_vecsize = 64;      /* we'll update this later */
	x->x_bitrate = 0;		/* not specified, use default */
	x->x_cbuf = NULL;
	x->x_blocksize = DEFAULT_AUDIO_BUFFER_SIZE;
	x->x_blockspersend = x->x_blocksize / x->x_vecsize;
	x->x_blockssincesend = 0;
	x->x_cbufsize = x->x_blocksize * sizeof(t_float) * x->x_ninlets;
	x->x_cbuf = (char *)t_getbytes(x->x_cbufsize);

#ifdef UNIX
	/* we don't want to get signaled in case send() fails */
	signal(SIGPIPE, SIG_IGN);
#endif


	for( i = 0; i < x->x_ninlets; i++ )
			x->x_tmpbuf_start[i] = getbytes(sizeof(t_float) * x->x_vecsize); 

	return (x);
}

/* --- nettest_TILDE_SETUP METHOD --- */
void nettest_tilde_setup(void)
{
	nettest_tilde_class = class_new(gensym("nettest~"), (t_newmethod)nettest_tilde_new, (t_method)nettest_tilde_free,
		sizeof(t_nettest_tilde), CLASS_DEFAULT, A_DEFFLOAT, A_NULL);
	class_addmethod(nettest_tilde_class, (t_method)nettest_tilde_dsp, gensym("dsp"), A_NULL);
	CLASS_MAINSIGNALIN(nettest_tilde_class, t_nettest_tilde, x_float);
	class_addfloat(nettest_tilde_class, nettest_tilde_float);
	class_addbang(nettest_tilde_class, nettest_tilde_bang);
	class_addmethod(nettest_tilde_class, (t_method)nettest_tilde_connect, gensym("connect"), A_DEFSYM, A_DEFFLOAT, 0);
	class_addmethod(nettest_tilde_class, (t_method)nettest_tilde_disconnect, gensym("disconnect"), 0);
	class_addmethod(nettest_tilde_class, (t_method)nettest_tilde_channels, gensym("channels"), A_FLOAT, 0);
	class_addmethod(nettest_tilde_class, (t_method)nettest_tilde_format, gensym("format"), A_SYMBOL, A_DEFFLOAT, 0);
	class_addmethod(nettest_tilde_class, (t_method)nettest_tilde_host, gensym("host"), A_DEFSYM, 0);
	/* rm */
	class_addmethod(nettest_tilde_class, (t_method)nettest_tilde_print, gensym("print"), 0);
	class_addmethod(nettest_tilde_class, (t_method)receive_kick, gensym("kick"), 0);
	class_addmethod(nettest_tilde_class, (t_method)receive_reset, gensym("reset"), A_DEFFLOAT, 0);
	class_addmethod(nettest_tilde_class, (t_method)receive_port, gensym("port"), A_DEFFLOAT, 0);
	class_sethelpsymbol(nettest_tilde_class, gensym("nettest~"));

	ps_nothing = gensym("");
	ps_localhost = gensym("localhost");
	ps_hostname = gensym("ipaddr");
	ps_format = gensym("format");
	ps_channels = gensym("channels");
	ps_framesize = gensym("framesize");
	ps_bitrate = gensym("bitrate");
	ps_sf_float = gensym("_float_");
	ps_sf_16bit = gensym("_16bit_");
	ps_sf_8bit = gensym("_8bit_");
	ps_sf_mp3 = gensym("_mp3_");
	ps_sf_aac = gensym("_aac_");
	ps_sf_unknown = gensym("_unknown_");
	/* rm */
	ps_overflow = gensym("overflow");
	ps_underflow = gensym("underflow");
	ps_queuesize = gensym("queuesize");
	ps_average = gensym("average");
}



/* --- NOTIFY CONNECTIONS ON OUTLET --- */
static void nettest_tilde_notify(t_nettest_tilde *x)
{
	pthread_mutex_lock(&x->x_mutex);
	x->x_childthread = 0;
	outlet_float(x->x_outlet, x->x_connectstate);
	pthread_mutex_unlock(&x->x_mutex);
}

/* --- DISCONNECT --- */
static void nettest_tilde_disconnect(t_nettest_tilde *x)
{
	pthread_mutex_lock(&x->x_mutex);
	if (x->x_fd != -1)
	{
		nettest_tilde_closesocket(x->x_fd);
		x->x_fd = -1;
		x->x_connectstate = 0;
		outlet_float(x->x_outlet, 0);
	}
	pthread_mutex_unlock(&x->x_mutex);
}



/* --- CONNECT CORE FUNCTION --- */
static void *nettest_tilde_doconnect(void *zz)
{
	t_nettest_tilde *x = (t_nettest_tilde *)zz;
	struct sockaddr_in server;
	struct hostent *hp;
	int intarg = 1;
	int sockfd;
	int portno;
	t_symbol *hostname;

	pthread_mutex_lock(&x->x_mutex);
	hostname = x->x_hostname;
	portno = x->x_portno;
	pthread_mutex_unlock(&x->x_mutex);

	/* create a socket */
	sockfd = socket(AF_INET, x->x_protocol, 0);
	if (sockfd < 0)
	{
		post("nettest~: connection to %s on port %d failed", hostname->s_name,portno);
		nettest_tilde_sockerror("socket");
		x->x_childthread = 0;
		return (0);
	}

	/* connect socket using hostname provided in command line */
	server.sin_family = AF_INET;
	hp = gethostbyname(x->x_hostname->s_name);
	if (hp == 0)
	{
		post("nettest~: bad host?");
		x->x_childthread = 0;
		return (0);
	}

	/* for stream (TCP) sockets, specify "nodelay" */
	if (x->x_protocol == SOCK_STREAM)
	{
		intarg = 1;
		if (setsockopt(sockfd, SOL_IP, TCP_NODELAY, (const char *)&intarg, sizeof(intarg)) < 0)
			error("nettest~: setsockopt(TCP_NODELAY) failed");
	}

#ifdef SO_PRIORITY
	/* set high priority, LINUX only */
	intarg = 6;	/* select a priority between 0 and 7 */
	if (setsockopt(sockfd, SOL_SOCKET, SO_PRIORITY, (const char*)&intarg, sizeof(int)) < 0)
	{
		error("nettest~: setsockopt(SO_PRIORITY) failed");
	}
#endif

	memcpy((char *)&server.sin_addr, (char *)hp->h_addr_list[0], hp->h_length);

	/* assign client port number */
	server.sin_port = htons((unsigned short)portno);

	/* try to connect */
	if (connect(sockfd, (struct sockaddr *) &server, sizeof (server)) < 0)
	{
        	nettest_tilde_sockerror("connecting stream socket");
        	nettest_tilde_closesocket(sockfd);
		x->x_childthread = 0;
		return (0);
	}

	post("nettest~: connected host %s on port %d", hostname->s_name, portno);

	pthread_mutex_lock(&x->x_mutex);
	x->x_fd = sockfd;
	x->x_connectstate = 1;
	clock_delay(x->x_clock, 0);
	pthread_mutex_unlock(&x->x_mutex);
	return (0);
}


/* --- CONNECT METHOD --- */
static void nettest_tilde_connect(t_nettest_tilde *x, t_symbol *host, t_floatarg fportno)
{
	pthread_mutex_lock(&x->x_mutex);
    	if (x->x_childthread != 0)
	{
		pthread_mutex_unlock(&x->x_mutex);
		post("nettest~: already trying to connect");
		return;
	}
	if (x->x_fd != -1)
	{
		pthread_mutex_unlock(&x->x_mutex);
		post("nettest~: already connected");
		return;
	}

	if (host != ps_nothing)
		x->x_hostname = host;
	else
		x->x_hostname = ps_localhost;

	if (!fportno)
		x->x_portno = DEFAULT_PORT;
	else
		x->x_portno = (int)fportno;
	x->x_count = 0;

	/* start child thread to connect */
	pthread_create(&x->x_childthread, 0, nettest_tilde_doconnect, x);
	pthread_mutex_unlock(&x->x_mutex);
}



#define QUEUESIZE (int)((x->x_framein + DEFAULT_AUDIO_BUFFER_FRAMES - x->x_frameout) % DEFAULT_AUDIO_BUFFER_FRAMES)
#define BLOCKOFFSET (x->x_blockssincerecv * x->x_nr_vecsize * x->x_frames[x->x_frameout].tag.channels)


void send_routine(t_int *w, const int offset) {
	t_nettest_tilde* x = (t_nettest_tilde*) (w[1]);
	int n = (int)(w[2]);
	t_float *in[DEFAULT_AUDIO_CHANNELS];
	char* bp = NULL;
	int i, length = x->x_blocksize * SF_SIZEOF(x->x_tag.format) * x->x_tag.channels;
	int sent = 0;
	pthread_mutex_lock(&x->x_mutex);

	for (i = 0; i < x->x_ninlets; i++) {
		in[i] = (t_float *)(w[offset + i]);
	}

	if (n != x->x_vecsize)	/* resize buffer */
	{
		post("nettest: Buffer resize");
		x->x_vecsize = n;
		x->x_blockspersend = x->x_blocksize / x->x_vecsize;
		x->x_blockssincesend = 0;
		length = x->x_blocksize * SF_SIZEOF(x->x_tag.format) * x->x_tag.channels;
	}

	/* format the buffer */
	switch (x->x_tag.format)
	{
		case SF_FLOAT:
		{
			t_float* fbuf = (t_float *)x->x_cbuf + (x->x_blockssincesend * x->x_vecsize * x->x_tag.channels);
			while( n-- > 0 )
				for (i = 0; i < x->x_tag.channels; i++)
					*fbuf++ = *(in[i]++);
			break;
		}
		case SF_16BIT:
		{
			short* cibuf = (short *)x->x_cbuf + (x->x_blockssincesend * x->x_vecsize * x->x_tag.channels);
			while( n-- > 0 )
				for (i = 0; i < x->x_tag.channels; i++)
					*cibuf++ = (short)lrint(32767.0 * *(in[i]++));
			break;
		}
		case SF_8BIT:
		{
			unsigned char*  cbuf = (unsigned char*)x->x_cbuf + (x->x_blockssincesend * x->x_vecsize * x->x_tag.channels);
			while( n-- > 0 )
				for (i = 0; i < x->x_tag.channels; i++)
					*cbuf++ = (unsigned char)(128. * (1.0 + *(in[i]++)));
			break;
		}
		default:
			 break;
	}

	if (!(x->x_blockssincesend < x->x_blockspersend - 1))	/* time to send the buffer */
	{
		x->x_blockssincesend = 0;
		x->x_count++;	/* count data packet we're going to send */

		if (x->x_fd != -1)
		{
			bp = (char *)x->x_cbuf;

			/* fill in the header tag */
			x->x_tag.framesize = length;

			/* CRUCIAL: NETSEND: adds a counter to the frame to send */
			x->x_tag.count = x->x_count;

			/* send the format tag */
			if (send(x->x_fd, (char*)&x->x_tag, sizeof(t_tag), SEND_FLAGS) < 0)
			{
				nettest_tilde_sockerror("send tag");
				pthread_mutex_unlock(&x->x_mutex);
				nettest_tilde_disconnect(x);
				return (w + offset + x->x_ninlets);
			}

			
			{ /* UDP: max. packet size is 64k (incl. headers) so we have to split */
#ifdef __APPLE__
				/* WARING: due to a 'bug' (maybe Apple would call it a feature?) in OS X
				   send calls with data packets larger than 16k fail with error number 40!
				   Thus we have to split the data packets into several packets that are
				   16k in size. The other side will reassemble them again. */

				int size = DEFAULT_UDP_PACKT_SIZE;
				if (length < size)	/* maybe data fits into one packet? */
					size = length;

				/* send the buffer */
				for (sent = 0; sent < length;)
				{
					int ret = 0;
					ret = send(x->x_fd, bp, size, SEND_FLAGS);
					if (ret <= 0)
					{
						nettest_tilde_sockerror("send data");
						pthread_mutex_unlock(&x->x_mutex);
						nettest_tilde_disconnect(x);
						return (w + offset + x->x_ninlets);
					}
					else
					{
						bp += ret;
						sent += ret;
						if ((length - sent) < size)
							size = length - sent;
					}
				}
#else
				/* send the buffer, the OS might segment it into smaller packets */
				int ret = send(x->x_fd, bp, length, SEND_FLAGS);
				if (ret <= 0)
				{
					nettest_tilde_sockerror("send data");
					pthread_mutex_unlock(&x->x_mutex);
					nettest_tilde_disconnect(x);
					return (w + offset + x->x_ninlets);
				}
#endif
			}
		}

		/* check whether user has updated any parameters */
		if (x->x_tag.channels != x->x_channels)
		{
			x->x_tag.channels = x->x_channels;
		}
		if (x->x_tag.format != x->x_format)
		{
			x->x_tag.format = x->x_format;
		}
	}
	else
	{
		x->x_blockssincesend++;
	}
	pthread_mutex_unlock(&x->x_mutex);
}



void receive_routine(t_int *w, const int offset)
{
	t_nettest_tilde *x = (t_nettest_tilde*) (w[1]);
	int n = (int)(w[2]);
	t_float *out[DEFAULT_AUDIO_CHANNELS];
	t_float *prev_out;
	const int channels = x->x_frames[x->x_frameout].tag.channels;
	int i = 0;
	int j = 0;


	pthread_mutex_lock(&x->x_mutex);

	for (i = 0; i < x->x_noutlets; i++)
	{
		out[i] = (t_float *)(w[offset + i]);

	}

	if (n != x->x_nr_vecsize)
	{
		post("nettest~: frame resize required");
		x->x_nr_vecsize = n;
		x->x_blocksperrecv = x->x_nr_blocksize / x->x_nr_vecsize;
		x->x_blockssincerecv = 0;
	}
	/* check whether there is enough data in buffer */
	if (x->x_counter < x->x_maxframes)
	{
		goto bail;
	}

	/* check for buffer underflow */
	if (x->x_framein == x->x_frameout)
	{
		x->x_underflow++;
#if DEBUG == 1
		if( x->x_underflow % 10 == 0)
			post("New buffer underflow: quota %d\n",x->x_underflow);
#endif
		goto bail;
	}


	nettest_latency_output(x);

	/* queue balancing */
	x->x_average[x->x_averagecur] = QUEUESIZE;
	if (++x->x_averagecur >= DEFAULT_AVERAGE_NUMBER)
		x->x_averagecur = 0;

	switch (x->x_frames[x->x_frameout].tag.format)
	{
		case SF_FLOAT:
		{
			t_float* buf = (t_float *)x->x_frames[x->x_frameout].data + BLOCKOFFSET;

			if (x->x_frames[x->x_frameout].tag.version == SF_BYTE_NATIVE)
			{
				while( n-- > 0 )
				{
					for (i = 0; i < channels; i++)
					{
						*out[i]++ = *buf++;
					}
					for (i = channels; i < x->x_noutlets; i++) 
					{
						*out[i]++ = 0.;
					}
				}
			}
			else	/* swap bytes */
			{
				while( n-- > 0 )
				{
					for (i = 0; i < channels; i++)
					{
						*out[i]++ = nettest_float(*buf++);
					}
					for (i = channels; i < x->x_noutlets; i++)
					{
						*out[i]++ = 0.;
					}
				}
			}
			break;
		}
		case SF_16BIT:
		{
			short* buf = (short *)x->x_frames[x->x_frameout].data + BLOCKOFFSET;

			if (x->x_frames[x->x_frameout].tag.version == SF_BYTE_NATIVE)
			{
				while( n-- > 0 )
				{
					for (i = 0; i < channels; i++)
					{
						*out[i]++ = (t_float)(*buf++ * 3.051850e-05);
					}
					for (i = channels; i < x->x_noutlets; i++)
					{
						*out[i]++ = 0.;
					}
				}
			}
			else /* swap bytes */
			{
				while( n-- > 0 )
				{
					for (i = 0; i < channels; i++)
					{
						*out[i]++ = (t_float)(nettest_short(*buf++) * 3.051850e-05);
					}
					for (i = channels; i < x->x_noutlets; i++)
					{
						*out[i]++ = 0.;
					}
				}
			}
			break;
		}
		case SF_8BIT:
		{
			unsigned char* buf = (char *)x->x_frames[x->x_frameout].data + BLOCKOFFSET;

			while( n-- > 0 )
			{
				for (i = 0; i < channels; i++)
				{
					*out[i]++ = (t_float)((0.0078125 * (*buf++)) - 1.0);
				}
				for (i = channels; i < x->x_noutlets; i++)
				{
					*out[i]++ = 0.;
				}
			}
			break;
		}
		default:
			post("netreceivev3~: unknown format (%d)",x->x_frames[x->x_frameout].tag.format);
			if (x->x_tcp)
				receive_kick(x);
			break;
	}
	if (!(x->x_blockssincerecv < x->x_blocksperrecv - 1))
	{
		x->x_blockssincerecv = 0;
		x->x_frameout++;
		x->x_frameout %= DEFAULT_AUDIO_BUFFER_FRAMES;
	}
	else
	{
		x->x_blockssincerecv++;
	}

	pthread_mutex_unlock(&x->x_mutex);


bail:
	/* set output to zero */
	while( n-- > 0 )
	{
		for (i = 0; i < x->x_noutlets; i++)
		{
			*(out[i]++) = 0.;
		}
	}
	pthread_mutex_unlock(&x->x_mutex);
}



/* --- PERFORM PD METHOD --- */
static t_int *nettest_tilde_perform(t_int *w) {

	const int offset = 3;
	t_nettest_tilde* x = (t_nettest_tilde*) (w[1]);

	/* offset+x_ninlets is needed to make netreceive get the right data pointers */
	/* i due forse dovrebbero girare in parallelo, cmq non vanno in race perché usano mutex */
	send_routine(w, offset);

	receive_routine(w, offset + x->x_ninlets);

    return (w + offset + x->x_ninlets + x->x_noutlets);
}



static t_int* sendback_routine(t_int* w) {

	const int offset = 3;
	t_nettest_tilde* x = (t_nettest_tilde*) (w[1]);

	t_signal asd;
	int n = (int)(w[2]);
	t_float *out[DEFAULT_AUDIO_CHANNELS];
	t_float* tmpbuf[DEFAULT_AUDIO_CHANNELS];
	const int channels = x->x_frames[x->x_frameout].tag.channels;
	int i = 0;
	int j = 0;

	pthread_mutex_lock(&x->x_mutex);

	for( i = 0; i < x->x_ninlets; i++ )
		tmpbuf[i] = x->x_tmpbuf_start[i];

	if (n != x->x_nr_vecsize)
	{
		post("nettest~: frame resize required");
		x->x_nr_vecsize = n;
		x->x_blocksperrecv = x->x_nr_blocksize / x->x_nr_vecsize;
		x->x_blockssincerecv = 0;
	}
	/* check whether there is enough data in buffer */
	if (x->x_counter < x->x_maxframes)
	{
		goto bail;
	}

	/* check for buffer underflow */
	if (x->x_framein == x->x_frameout)
	{
		x->x_underflow++;
#if DEBUG == 1
		if( x->x_underflow % 10 == 0)
			post("New buffer underflow: quota %d\n",x->x_underflow);
#endif
		goto bail;
	}


	/* queue balancing */
	x->x_average[x->x_averagecur] = QUEUESIZE;
	if (++x->x_averagecur >= DEFAULT_AVERAGE_NUMBER)
		x->x_averagecur = 0;

	switch (x->x_frames[x->x_frameout].tag.format)
	{
		case SF_FLOAT:
		{
			t_float* buf = (t_float *)x->x_frames[x->x_frameout].data + BLOCKOFFSET;

			if (x->x_frames[x->x_frameout].tag.version == SF_BYTE_NATIVE)
			{
				while( n-- > 0 )
				{
					for (i = 0; i < channels; i++)
					{
						*tmpbuf[i]++ = *buf++;
					}
					for (i = channels; i < x->x_noutlets; i++) 
					{
						*tmpbuf[i]++ = 0.;
					}
				}
			}
			else	/* swap bytes */
			{
				while( n-- > 0 )
				{
					for (i = 0; i < channels; i++)
					{
						*tmpbuf[i]++ = nettest_float(*buf++);
					}
					for (i = channels; i < x->x_noutlets; i++)
					{
						*tmpbuf[i]++ = 0.;
					}
				}
			}
			break;
		}
		case SF_16BIT:
		{
			short* buf = (short *)x->x_frames[x->x_frameout].data + BLOCKOFFSET;

			if (x->x_frames[x->x_frameout].tag.version == SF_BYTE_NATIVE)
			{
				while( n-- > 0 )
				{
					for (i = 0; i < channels; i++)
					{
						*tmpbuf[i]++ = (t_float)(*buf++ * 3.051850e-05);
					}
					for (i = channels; i < x->x_noutlets; i++)
					{
						*tmpbuf[i]++ = 0.;
					}
				}
			}
			else /* swap bytes */
			{
				while( n-- > 0 )
				{
					for (i = 0; i < channels; i++)
					{
						*tmpbuf[i]++ = (t_float)(nettest_short(*buf++) * 3.051850e-05);
					}
					for (i = channels; i < x->x_noutlets; i++)
					{
						*tmpbuf[i]++ = 0.;
					}
				}
			}
			break;
		}
		case SF_8BIT:
		{
			unsigned char* buf = (char *)x->x_frames[x->x_frameout].data + BLOCKOFFSET;

			while( n-- > 0 )
			{
				for (i = 0; i < channels; i++)
				{
					*tmpbuf[i]++ = (t_float)((0.0078125 * (*buf++)) - 1.0);
				}
				for (i = channels; i < x->x_noutlets; i++)
				{
					*tmpbuf[i]++ = 0.;
				}
			}
			break;
		}
		default:
			post("netreceivev3~: unknown format (%d)",x->x_frames[x->x_frameout].tag.format);
			if (x->x_tcp)
				receive_kick(x);
			break;
	}
	if (!(x->x_blockssincerecv < x->x_blocksperrecv - 1))
	{
		x->x_blockssincerecv = 0;
		x->x_frameout++;
		x->x_frameout %= DEFAULT_AUDIO_BUFFER_FRAMES;
	}
	else
	{
		x->x_blockssincerecv++;
	}

	pthread_mutex_unlock(&x->x_mutex);


	nettest_latency_output(x);


	/* adapted from send_routine */

	t_float *in[DEFAULT_AUDIO_CHANNELS];
	char* bp = NULL;
	int length = x->x_blocksize * SF_SIZEOF(x->x_tag.format) * x->x_tag.channels;
	int sent = 0;
	n = (int)(w[2]);

	pthread_mutex_lock(&x->x_mutex);

	for (i = 0; i < x->x_ninlets; i++) {
		in[i] = x->x_tmpbuf_start[i];
	}

	if (n != x->x_vecsize)	/* resize buffer */
	{
		post("nettest: Buffer resize");
		x->x_vecsize = n;
		x->x_blockspersend = x->x_blocksize / x->x_vecsize;
		x->x_blockssincesend = 0;
		length = x->x_blocksize * SF_SIZEOF(x->x_tag.format) * x->x_tag.channels;
	}

	/* format the buffer */
	switch (x->x_tag.format)
	{
	case SF_FLOAT:
	{
		t_float* fbuf = (t_float *)x->x_cbuf + (x->x_blockssincesend * x->x_vecsize * x->x_tag.channels);
		while( n-- > 0 )
			for (i = 0; i < x->x_tag.channels; i++)
				*fbuf++ = *(in[i]++);
		break;
	}
	case SF_16BIT:
	{
		short* cibuf = (short *)x->x_cbuf + (x->x_blockssincesend * x->x_vecsize * x->x_tag.channels);
		while( n-- > 0 )
			for (i = 0; i < x->x_tag.channels; i++)
				*cibuf++ = (short)lrint(32767.0 * *(in[i]++));
		break;
	}
	case SF_8BIT:
	{
		unsigned char*  cbuf = (unsigned char*)x->x_cbuf + (x->x_blockssincesend * x->x_vecsize * x->x_tag.channels);
		while( n-- > 0 )
			for (i = 0; i < x->x_tag.channels; i++)
				*cbuf++ = (unsigned char)(128. * (1.0 + *(in[i]++)));
		break;
	}
	default:
		break;
	}

	if (!(x->x_blockssincesend < x->x_blockspersend - 1))	/* time to send the buffer */
	{
		x->x_blockssincesend = 0;
		x->x_count++;	/* count data packet we're going to send */

		if (x->x_fd != -1)
		{
			bp = (char *)x->x_cbuf;

			/* fill in the header tag */
			x->x_tag.framesize = length;

			/* CRUCIAL: SENDBACK: copies the received count on the frame to send */
			x->x_tag.count = x->x_frames[x->x_framein].tag.count;

			/* send the format tag */
			if (send(x->x_fd, (char*)&x->x_tag, sizeof(t_tag), SEND_FLAGS) < 0)
			{
				nettest_tilde_sockerror("send tag");
				pthread_mutex_unlock(&x->x_mutex);
				nettest_tilde_disconnect(x);
				return (w + offset + x->x_noutlets + x->x_ninlets);
			}

			
			{ /* UDP: max. packet size is 64k (incl. headers) so we have to split */
#ifdef __APPLE__
				/* WARING: due to a 'bug' (maybe Apple would call it a feature?) in OS X
					   send calls with data packets larger than 16k fail with error number 40!
					   Thus we have to split the data packets into several packets that are
					   16k in size. The other side will reassemble them again. */

				int size = DEFAULT_UDP_PACKT_SIZE;
				if (length < size)	/* maybe data fits into one packet? */
					size = length;

				/* send the buffer */
				for (sent = 0; sent < length;)
				{
					int ret = 0;
					ret = send(x->x_fd, bp, size, SEND_FLAGS);
					if (ret <= 0)
					{
						nettest_tilde_sockerror("send data");
						pthread_mutex_unlock(&x->x_mutex);
						nettest_tilde_disconnect(x);
						return (w + offset + x->x_noutlets + x->x_ninlets);
					}
					else
					{
						bp += ret;
						sent += ret;
						if ((length - sent) < size)
							size = length - sent;
					}
				}
#else
				/* send the buffer, the OS might segment it into smaller packets */
				int ret = send(x->x_fd, bp, length, SEND_FLAGS);
				if (ret <= 0)
				{
					nettest_tilde_sockerror("send data");
					pthread_mutex_unlock(&x->x_mutex);
					nettest_tilde_disconnect(x);
					return (w + offset + x->x_noutlets + x->x_ninlets);
				}
#endif
			}
		}

		/* check whether user has updated any parameters */
		if (x->x_tag.channels != x->x_channels)
		{
			x->x_tag.channels = x->x_channels;
		}
		if (x->x_tag.format != x->x_format)
		{
			x->x_tag.format = x->x_format;
		}
	}
	else
	{
		x->x_blockssincesend++;
	}
	pthread_mutex_unlock(&x->x_mutex);

	return (w + offset + x->x_noutlets + x->x_ninlets);

bail:
	/* set output to zero */
	while( n-- > 0 )
	{
		for (i = 0; i < x->x_noutlets; i++)
		{
			*(tmpbuf[i]++) = 0.;
		}
	}
	pthread_mutex_unlock(&x->x_mutex);
	return (w + offset + x->x_noutlets + x->x_ninlets);

}


/* --- DSP METHOD --- */
static void nettest_tilde_dsp(t_nettest_tilde *x, t_signal **sp)
{
	int i;

	pthread_mutex_lock(&x->x_mutex);
	x->x_myvec[0] = (t_int*)x;
	x->x_myvec[1] = (t_int*)sp[0]->s_n;
	x->x_samplerate = sp[0]->s_sr;


	for (i = 0; i < x->x_ninlets; i++)
		x->x_myvec[2 + i] = (t_int*)sp[i]->s_vec;

	pthread_mutex_unlock(&x->x_mutex);

	if (DEFAULT_AUDIO_BUFFER_SIZE % sp[0]->s_n)
	{
		error("nettest~: signal vector size too large (needs to be even divisor of %d)", DEFAULT_AUDIO_BUFFER_SIZE);
	}
	else
	{
		pthread_mutex_lock(&x->x_mutex);
		for (i = x->x_ninlets; i < (x->x_ninlets + x->x_noutlets); i++)
			x->x_myvec[2 + i] = (t_int*)sp[i]->s_vec;

		pthread_mutex_unlock(&x->x_mutex);
		if(x->x_mode == 1) {
			dsp_addv(nettest_tilde_perform, x->x_ninlets + x->x_noutlets + 2, (t_int*)x->x_myvec);
		}
		else if(x->x_mode == 2)
			dsp_addv(sendback_routine, x->x_ninlets + x->x_noutlets + 2, (t_int*)x->x_myvec);
		else
			error("nettest~: creation argument can be either 1 or 2");
	}
}


static void nettest_tilde_channels(t_nettest_tilde *x, t_floatarg channels)
{
	pthread_mutex_lock(&x->x_mutex);
	if (channels >= 0 && channels <= DEFAULT_AUDIO_CHANNELS)
	{
		x->x_channels = (int)channels;
		post("nettest~: channels set to %d", (int)channels);
	}
	pthread_mutex_unlock(&x->x_mutex);
}



static void nettest_tilde_format(t_nettest_tilde *x, t_symbol* form, t_floatarg bitrate)
{
	pthread_mutex_lock(&x->x_mutex);
	if (!strncmp(form->s_name,"float", 5) && x->x_tag.format != SF_FLOAT)
	{
		x->x_format = (int)SF_FLOAT;
	}
	else if (!strncmp(form->s_name,"16bit", 5) && x->x_tag.format != SF_16BIT)
	{
		x->x_format = (int)SF_16BIT;
	}
	else if (!strncmp(form->s_name,"8bit", 4) && x->x_tag.format != SF_8BIT)
	{
		x->x_format = (int)SF_8BIT;
	}
	post("nettest~: format set to %s", form->s_name);
	pthread_mutex_unlock(&x->x_mutex);
}


/* set hostname to send to */
static void nettest_tilde_host(t_nettest_tilde *x, t_symbol* host)
{
	pthread_mutex_lock(&x->x_mutex);
	if (host != ps_nothing)
		x->x_hostname = host;
	else
		x->x_hostname = ps_localhost;

	if (x->x_fd != -1)
	{
		pthread_mutex_unlock(&x->x_mutex);
		nettest_tilde_connect(x,x->x_hostname, (float)x->x_portno);
		return;
	}
	pthread_mutex_unlock(&x->x_mutex);
}



static void nettest_tilde_float(t_nettest_tilde* x, t_floatarg arg)
{
	if (arg == 0.0)
		nettest_tilde_disconnect(x);
	else
		nettest_tilde_connect(x,x->x_hostname,(float) x->x_portno);
}


/* send stream info when banged */
static void nettest_tilde_bang(t_nettest_tilde *x)
{
	t_atom list[2];
	t_symbol *sf_format;
	t_float bitrate;
	int i, avg = 0;
	for (i = 0; i < DEFAULT_AVERAGE_NUMBER; i++)
		avg += x->x_average[i];

	bitrate = (t_float)((SF_SIZEOF(x->x_tag.format) * x->x_samplerate * 8 * x->x_tag.channels) / 1000.);

	switch (x->x_tag.format)
	{
		case SF_FLOAT:
		{
			sf_format = ps_sf_float;
			break;
		}
		case SF_16BIT:
		{
			sf_format = ps_sf_16bit;
			break;
		}
		case SF_8BIT:
		{
			sf_format = ps_sf_8bit;
			break;
		}
		default:
		{
			sf_format = ps_sf_unknown;
			break;
		}
	}

	/* --- stream information (t_tag) --- */
	/* audio format */
	SETSYMBOL(list, (t_symbol *)sf_format);
	outlet_anything(x->x_outlet2, ps_format, 1, list);

	/* channels */
	SETFLOAT(list, (t_float)x->x_tag.channels);
	outlet_anything(x->x_outlet2, ps_channels, 1, list);

	/* framesize */
	SETFLOAT(list, (t_float)x->x_tag.framesize);
	outlet_anything(x->x_outlet2, ps_framesize, 1, list);

	/* bitrate */
	SETFLOAT(list, (t_float)bitrate);
	outlet_anything(x->x_outlet2, ps_bitrate, 1, list);

	/* --- internal info (buffer and network) --- */
	/* overflow */
	SETFLOAT(list, (t_float)x->x_overflow);
	outlet_anything(x->x_outlet4, ps_overflow, 1, list);

	/* underflow */
	SETFLOAT(list, (t_float)x->x_underflow);
	outlet_anything(x->x_outlet4, ps_underflow, 1, list);

	/* queuesize */
	SETFLOAT(list, (t_float)QUEUESIZE);
	outlet_anything(x->x_outlet4, ps_queuesize, 1, list);

	/* average queuesize */
	SETFLOAT(list, (t_float)((t_float)avg / (t_float)DEFAULT_AVERAGE_NUMBER));
	outlet_anything(x->x_outlet4, ps_average, 1, list);

	/* IP address */
	SETSYMBOL(list, (t_symbol *)x->x_hostname);
	outlet_anything(x->x_outlet2, ps_hostname, 1, list);

}





static void nettest_tilde_free(t_nettest_tilde* x)
{
	nettest_tilde_disconnect(x);

	/* free the memory */
	if (x->x_cbuf)t_freebytes(x->x_cbuf, x->x_cbufsize);
	if (x->x_myvec)t_freebytes(x->x_myvec, sizeof(t_int) * (x->x_ninlets + 3));

	clock_free(x->x_clock);

	pthread_cond_destroy(&x->x_requestcondition);
	pthread_cond_destroy(&x->x_answercondition);
	pthread_mutex_destroy(&x->x_mutex);
    
	/* rm */
	int i;

	if (x->x_connectsocket != -1)
	{
		sys_rmpollfn(x->x_connectsocket);
		CLOSESOCKET(x->x_connectsocket);
	}
	if (x->x_socket != -1)
	{
		sys_rmpollfn(x->x_socket);
		CLOSESOCKET(x->x_socket);
	}

	/* free memory */
	for (i = 0; i < DEFAULT_AUDIO_BUFFER_FRAMES; i++)
	{
		 t_freebytes(x->x_frames[i].data, DEFAULT_AUDIO_BUFFER_SIZE * x->x_noutlets * sizeof(t_float));
	}
}







/* --- UTILITY FUNCTIONS --- */
static int nettest_tilde_sockerror(char *s)
{
#ifdef _WINDOWS
    int err = WSAGetLastError();
    if (err == 10054) return 1;
    else if (err == 10053) post("nettest~: %s: software caused connection abort (%d)", s, err);
    else if (err == 10055) post("nettest~: %s: no buffer space available (%d)", s, err);
    else if (err == 10060) post("nettest~: %s: connection timed out (%d)", s, err);
    else if (err == 10061) post("nettest~: %s: connection refused (%d)", s, err);
    else post("nettest~: %s: %s (%d)", s, strerror(err), err);
#else
    int err = errno;
    post("nettest~: %s: %s (%d)", s, strerror(err), err);
#endif
#ifdef _WINDOWS
	if (err == WSAEWOULDBLOCK)
#endif
#ifdef UNIX
	if (err == EAGAIN)
#endif
	{
		return 1;	/* recoverable error */
	}
	return 0;	/* indicate non-recoverable error */
}



static void nettest_tilde_closesocket(int fd)
{
#ifdef UNIX
	close(fd);
#endif
#ifdef NT
	closesocket(fd);
#endif
}








/* remove all pollfunctions and close socket */
static void receive_closesocket(t_nettest_tilde* x)
{
	sys_rmpollfn(x->x_socket);
	outlet_float(x->x_outlet3, 0);
	CLOSESOCKET(x->x_socket);
	x->x_socket = -1;
}



static void receive_reset(t_nettest_tilde* x, t_floatarg buffer)
{
	int i;
	x->x_counter = 0;
	x->x_nbytes = 0;
	x->x_framein = 0;
	x->x_frameout = 0;
	x->x_blockssincerecv = 0;
	x->x_blocksperrecv = x->x_nr_blocksize / x->x_nr_vecsize;

	for (i = 0; i < DEFAULT_AVERAGE_NUMBER; i++)
		x->x_average[i] = x->x_maxframes;
	x->x_averagecur = 0;

	if (buffer == 0.0)	/* set default */
		x->x_maxframes = DEFAULT_QUEUE_LENGTH;
	else
	{
		buffer = (float)CLIP((float)buffer, 0., 1.);
		x->x_maxframes = (int)(DEFAULT_AUDIO_BUFFER_FRAMES * buffer);
		x->x_maxframes = CLIP(x->x_maxframes, 1, DEFAULT_AUDIO_BUFFER_FRAMES - 1);
		post("netreceivev3~: set buffer to %g (%d frames)", buffer, x->x_maxframes);
	}
	x->x_underflow = 0;
	x->x_overflow = 0;
}


static void receive_datapoll(t_nettest_tilde *x)
{
#ifndef PD
	int ret;
    struct timeval timout;
    fd_set readset;
    timout.tv_sec = 0;
    timout.tv_usec = 0;
    FD_ZERO(&readset);
    FD_SET(x->x_socket, &readset);

	ret = select(x->x_socket + 1, &readset, NULL, NULL, &timout);
    if (ret < 0)
    {
    	receive_sockerror("select");
		return;
    }

	if (FD_ISSET(x->x_socket, &readset))	/* data available */
#endif
	{
		int ret;
		int n;

		if (x->x_tcp)
		{
			n = x->x_nbytes;

			if (x->x_nbytes == 0)	/* we ate all the samples and need a new header tag */
			{
				/* get the new tag */
				ret = recv(x->x_socket, (char*)&x->x_frames[x->x_framein].tag, sizeof(t_tag), MSG_PEEK);
				if (ret == 0)	/* disconnect */
				{
					post("netreceivev3~: EOF on socket %d", x->x_socket);
					receive_closesocket(x);
					x->x_socket = -1;
					x->x_counter = 0;
					return;
				}
				if (ret < 0)	/* error */
				{
					if (receive_sockerror("recv tag"))
						goto bail;
					receive_closesocket(x);
					x->x_socket = -1;
					x->x_counter = 0;
					return;
				}
				else if (ret != sizeof(t_tag))
				{
					/* incomplete header tag: return and try again later */
					/* in the hope that more data will be available */
					return;
				}

				/* receive header tag */
				ret = recv(x->x_socket, (char*)&x->x_frames[x->x_framein].tag, sizeof(t_tag), 0);

				/* adjust byte order if neccessarry */
				if (x->x_frames[x->x_framein].tag.version != SF_BYTE_NATIVE)
				{
					x->x_frames[x->x_framein].tag.count = nettest_long(x->x_frames[x->x_framein].tag.count);
					x->x_frames[x->x_framein].tag.framesize = nettest_long(x->x_frames[x->x_framein].tag.framesize);
				}

				/* get info from header tag */
				if (x->x_frames[x->x_framein].tag.channels > x->x_noutlets)
				{
					error("nettest~: incoming stream has too many channels (%d), kicking client", x->x_frames[x->x_framein].tag.channels);
					receive_kick(x);
					x->x_socket = -1;
					x->x_counter = 0;
					return;
				}
				x->x_nbytes = n = x->x_frames[x->x_framein].tag.framesize;

				/* check whether the data packet has the correct count */
				if ((x->x_framecount != x->x_frames[x->x_framein].tag.count)
					&& (x->x_frames[x->x_framein].tag.count > 2))
				{
					error("netreceivev3~: we lost %d frames", (int)(x->x_frames[x->x_framein].tag.count - x->x_framecount));
					post("netreceivev3~: current package is %d, expected %d", x->x_frames[x->x_framein].tag.count, x->x_framecount);
				}
				x->x_framecount = x->x_frames[x->x_framein].tag.count + 1;
			}
			else	/* we already have the header tag or some data and need more */
			{
				ret = recv(x->x_socket, (char*)x->x_frames[x->x_framein].data + x->x_frames[x->x_framein].tag.framesize - n, n, 0);
				if (ret > 0)
				{
					n -= ret;
				}
				else if (ret < 0)	/* error */
				{
					if (receive_sockerror("recv data"))
						goto bail;
					receive_closesocket(x);
					x->x_socket = -1;
					x->x_counter = 0;
					return;
				}

				x->x_nbytes = n;
				if (n == 0)			/* a complete packet is received */
				{

					x->x_counter++;
					x->x_framein++;
					x->x_framein %= DEFAULT_AUDIO_BUFFER_FRAMES;

					/* check for buffer overflow */
					if (x->x_framein == x->x_frameout)
					{
						x->x_overflow++;
						if( x->x_overflow % 10 == 0)
							post("New buffer overflow: quota %d\n",x->x_overflow);
					}
				}
			}
		}
		else /* UDP */
		{
			n = x->x_nbytes;

			if (x->x_nbytes == 0)	/* we ate all the samples and need a new header tag */
			{
				/* receive header tag */
				ret = recv(x->x_socket, (char*)&x->x_frames[x->x_framein].tag, sizeof(t_tag), 0);
				if (ret <= 0)	/* error */
				{
					if (receive_sockerror("recv tag"))
						goto bail;
					receive_reset(x, 0);
					x->x_counter = 0;
					return;
				}
				else if (ret != sizeof(t_tag))
				{
					/* incomplete header tag: return and try again later */
					/* in the hope that more data will be available */
					error("netreceivev3~: got incomplete header tag");
					return;
				}
				/* adjust byte order if neccessarry */
				if (x->x_frames[x->x_framein].tag.version != SF_BYTE_NATIVE)
				{
					x->x_frames[x->x_framein].tag.count = nettest_long(x->x_frames[x->x_framein].tag.count);
					x->x_frames[x->x_framein].tag.framesize = nettest_long(x->x_frames[x->x_framein].tag.framesize);
				}
				/* get info from header tag */
				if (x->x_frames[x->x_framein].tag.channels > x->x_noutlets)
				{
					error("nettest~: incoming stream has too many channels (%d)", x->x_frames[x->x_framein].tag.channels);
					x->x_counter = 0;
					return;
				}
				x->x_nbytes = n = x->x_frames[x->x_framein].tag.framesize;
			}
			else	/* we already have header tag or some data and need more */
			{
				ret = recv(x->x_socket, (char*)x->x_frames[x->x_framein].data + x->x_frames[x->x_framein].tag.framesize - n, n, 0);
				if (ret > 0)
				{
					n -= ret;
				}
				else if (ret < 0)	/* error */
				{
					if (receive_sockerror("recv data"))
						goto bail;
					receive_reset(x, 0);
					x->x_counter = 0;
					return;
				}

				x->x_nbytes = n;
				if (n == 0)			/* a complete packet is received */
				{
					x->x_counter++;
					x->x_framein++;
					x->x_framein %= DEFAULT_AUDIO_BUFFER_FRAMES;

					/* check for buffer overflow */
					if (x->x_framein == x->x_frameout)
					{
						x->x_overflow++;
						if( x->x_overflow % 10 == 0)
							post("New buffer overflow: quota %d\n",x->x_overflow);
					}
				}
			}
		}
		/* output the latency calculated based on header tags */
		nettest_latency_calc(x, x->x_frames[x->x_framein].tag.count, x->x_count); // if mode=2 latency should be zero: check for debug
	}
bail:
	;
#ifndef PD
	clock_delay(x->x_datapoll, DEFAULT_NETWORK_POLLTIME);
#endif
}


static void receive_connectpoll(t_nettest_tilde *x)
{

		int sockaddrl = (int)sizeof(struct sockaddr);
		struct sockaddr_in incomer_address;
		int fd = accept(x->x_connectsocket, (struct sockaddr*)&incomer_address, &sockaddrl);
		if (fd < 0)
		{
			post("netreceivev3~: accept failed");
			return;
		}
#ifdef O_NONBLOCK
		fcntl(fd, F_SETFL, O_NONBLOCK);
#endif
		if (x->x_socket != -1)
		{
			post("netreceivev3~: new connection");
			receive_closesocket(x);
		}

		receive_reset(x, 0);
		x->x_socket = fd;
		x->x_nbytes = 0;
		x->x_nr_hostname = gensym(inet_ntoa(incomer_address.sin_addr));
		sys_addpollfn(fd, receive_datapoll, x);
		outlet_float(x->x_outlet3, 1);

#ifndef PD
	clock_delay(x->x_connectpoll, DEFAULT_NETWORK_POLLTIME);
#endif
}


static int receive_createsocket(t_nettest_tilde* x, int portno)
{
    struct sockaddr_in server;
    int sockfd;
    int tcp = x->x_tcp;

     /* create a socket */
    if (!tcp) {
      sockfd = socket(AF_INET, SOCK_DGRAM, 0);
      post("nettest: creating UDP socket!");
    }
    else {
      sockfd = socket(AF_INET, SOCK_STREAM, 0);
      post("nettest: creating a TCP socket! (UDP recommended: put 3rd argument = 1)");
    }

    if (sockfd < 0)
    {
        receive_sockerror("socket");
        return 0;
    }
    server.sin_family = AF_INET;
    server.sin_addr.s_addr = INADDR_ANY;

    /* assign server port number */

    server.sin_port = htons((u_short)portno);
    post("listening to port number %d", portno);

    receive_setsocketoptions(sockfd);

    /* name the socket */
    if (bind(sockfd, (struct sockaddr *)&server, sizeof(server)) < 0)
	{
         receive_sockerror("bind");
         CLOSESOCKET(sockfd);
         return 0;
    }


    if (!tcp)
	{
         x->x_socket = sockfd;
         x->x_nbytes = 0;
#ifdef PD
         sys_addpollfn(sockfd, receive_datapoll, x); // questa funzione viene aggiunta alla creazione del socket... è un problema se il socket non viene creato all'avvio?
#else
		 clock_delay(x->x_datapoll, 0);
#endif
    }
    else
	{
         if (listen(sockfd, 5) < 0)
		 {
              receive_sockerror("listen");
              CLOSESOCKET(sockfd);
			  return 0;
         }
         else
		 {
              x->x_connectsocket = sockfd;
			  /* start polling for connection requests */
#ifdef PD
              sys_addpollfn(sockfd, receive_connectpoll, x);
#else
			  clock_delay(x->x_connectpoll, 0);
#endif
         }
    }
    return 1;
}



/* kick connected client */
static void receive_kick(t_nettest_tilde *x)
{
	if (x->x_tcp)
	{
		if (x->x_socket != -1)
		{
			shutdown(x->x_socket, 1);
			receive_closesocket(x);
			post("netreceivev3~: kicked client!");
		}
		else error("netreceivev3~: no client to kick");
	}
	else error("netreceivev3~: kicking clients in UDP mode not possible");
}






static void nettest_tilde_print(t_nettest_tilde* x)
{
	int i, avg = 0;
	for (i = 0; i < DEFAULT_AVERAGE_NUMBER; i++)
		avg += x->x_average[i];
	post("nettest~: last size = %d, avg size = %g, %d underflows, %d overflows", QUEUESIZE, (float)((float)avg / (float)DEFAULT_AVERAGE_NUMBER), x->x_underflow, x->x_overflow);
	post("nettest~: channels = %d, framesize = %d, packets = %d", x->x_frames[x->x_framein].tag.channels, x->x_frames[x->x_framein].tag.framesize, x->x_counter);
	post("nettest~: LATENCY = %g", x->x_latency);
}




static void receive_port(t_nettest_tilde* x, t_floatarg fportno) {

	if (!receive_createsocket(x, (int)fportno))
	{
		error("netreceivev3~: failed to create listening socket");
	}

}











void nettest_latency_calc(t_nettest_tilde* x, int nr_count, int ns_count) {

	if( x->x_samplerate ) {
		if( ns_count > nr_count )
			x->x_latency = (t_float)((t_float)((ns_count - nr_count) * x->x_vecsize)  / x->x_samplerate); // ms
		else
			x->x_latency = -0.1;
	}
	else
		x->x_latency = -0.2;
}




void nettest_latency_output(t_nettest_tilde* x) {

	outlet_float(x->x_outlet_lat, x->x_latency);
	outlet_float(x->x_outlet3, x->x_frames[x->x_framein].tag.count);
	outlet_float(x->x_outlet4, x->x_count);
}
